import re

SPAMC_CONFIG_FILE = '/etc/mail/spamassassin/spamc.conf'
SPAMASSASSIN_SERVICE_OVERRIDE = '/etc/systemd/system/spamassassin.service'
SYSCONFIG_SPAMASSASSIN = '/etc/sysconfig/spamassassin'
SYSCONFIG_VARIABLE = 'SPAMDOPTIONS'
SPAMD_SHORTOPTS_NOARG = "ch46LlxPQqVv"
""" All short options in spamd that do not accept an argument, excluding -d. """


def parse_sysconfig_spamassassin(content):
    """
    Splits up a spamassassin sysconfig file into three parts and returns those parts:
    1. Beginning of the file up to the SPAMDOPTIONS assignment
    2. The assignment to the SPAMDOPTIONS variable (this is the assignment
       that takes effect, i.e. the last assignment to the variable)
    3. End of the file after the SPAMDOPTIONS assignment
    """
    line_continues = False
    is_assignment = False
    assignment_start = None
    assignment_end = None
    lines = content.split('\n')
    for ix, line in enumerate(lines):
        is_assignment = ((is_assignment and line_continues) or
                         (not (not is_assignment and line_continues) and
                          re.match(r'\s*' + SYSCONFIG_VARIABLE + '=', line)))
        if is_assignment:
            if line_continues:
                assignment_end += 1
            else:
                assignment_start = ix
                assignment_end = ix + 1
        line_continues = line.endswith('\\')

    if assignment_start is None:
        return content, '', ''
    assignment = ''
    for line in lines[assignment_start:assignment_end - 1]:
        assignment += line[:-1]
    assignment += lines[assignment_end - 1]
    pre_assignment = '\n'.join(lines[:assignment_start])
    post_assignment = '\n'.join(lines[assignment_end:])
    return pre_assignment, assignment, post_assignment



from leapp.libraries.actor.private_firewalldcollectusedobjectnames import (
    get_used_services,
    get_used_services_tuple,
    is_policy_in_use,
    is_zone_in_use,
    is_zone_in_use_tuple
)


def test_is_zone_in_use():
    conf = {'interfaces': ['dummy0'],
            'services': ['tftp-client']}
    assert is_zone_in_use(conf)

    conf = {'sources': ['10.1.2.0/24'],
            'services': ['tftp-client']}
    assert is_zone_in_use(conf)

    conf = {'interfaces': ['dummy0'],
            'sources': ['fd00::/8'],
            'services': ['tftp-client']}
    assert is_zone_in_use(conf)


def test_is_zone_in_use_tuple():
    conf = (None, None, None, None, None,
            ['tftp-client'],  # conf[5], services
            None, None, None, None,
            ['dummy0'],  # conf[10], interfaces
            [],  # conf[11], sources
            [],  # conf[12], rules_str
            None, None, None)
    assert is_zone_in_use_tuple(conf)

    conf = (None, None, None, None, None,
            ['tftp-client'],  # conf[5], services
            None, None, None, None,
            [],  # conf[10], interfaces
            ['10.1.2.0/24'],  # conf[11], sources
            [],  # conf[12], rules_str
            None, None, None)
    assert is_zone_in_use_tuple(conf)

    conf = (None, None, None, None, None,
            ['tftp-client'],  # conf[5], services
            None, None, None, None,
            ['dummy0'],  # conf[10], interfaces
            ['fd00::/8'],  # conf[11], sources
            [],  # conf[12], rules_str
            None, None, None)
    assert is_zone_in_use_tuple(conf)


def test_is_zone_in_use_negative():
    conf = {'interfaces': [],
            'services': ['tftp-client']}
    assert not is_zone_in_use(conf)

    conf = {'sources': [],
            'services': ['tftp-client']}
    assert not is_zone_in_use(conf)

    conf = {'services': ['tftp-client']}
    assert not is_zone_in_use(conf)


def test_is_zone_in_use_tuple_negative():
    conf = (None, None, None, None, None,
            ['tftp-client'],  # conf[5], services
            None, None, None, None,
            [],  # conf[10], interfaces
            [],  # conf[11], sources
            [],  # conf[12], rules_str
            None, None, None)
    assert not is_zone_in_use_tuple(conf)


def test_is_policy_in_use():
    conf = {'ingress_zones': ['HOST'],
            'egress_zones': ['public'],
            'services': ['tftp-client']}
    used_zones = ['public']
    assert is_policy_in_use(conf, used_zones)

    conf = {'ingress_zones': ['internal'],
            'egress_zones': ['external'],
            'services': []}
    used_zones = ['internal', 'external']
    assert is_policy_in_use(conf, used_zones)

    conf = {'ingress_zones': ['internal'],
            'egress_zones': ['external'],
            'services': []}
    used_zones = ['internal', 'external', 'public']
    assert is_policy_in_use(conf, used_zones)


def test_is_policy_in_use_negative():
    conf = {'ingress_zones': ['HOST'],
            'egress_zones': ['public'],
            'services': ['tftp-client']}
    used_zones = ['home']
    assert not is_policy_in_use(conf, used_zones)

    conf = {'ingress_zones': ['internal'],
            'egress_zones': ['external'],
            'services': []}
    used_zones = ['public', 'external']
    assert not is_policy_in_use(conf, used_zones)

    conf = {'egress_zones': ['external'],
            'services': []}
    used_zones = ['internal', 'public']
    assert not is_policy_in_use(conf, used_zones)


def test_get_used_services_zone():
    conf = {'interfaces': ['dummy0'],
            'services': ['tftp-client']}
    assert 'tftp-client' in get_used_services(conf, True)

    conf = {'sources': ['10.1.2.0/24'],
            'rules_str': ['rule family="ipv4" source address="10.1.1.0/24" service name="tftp-client" reject']}
    assert 'tftp-client' in get_used_services(conf, True)

    conf = {'interfaces': ['dummy0'],
            'sources': ['fd00::/8'],
            'rules_str': ['rule service name="ssh" accept',
                          'rule service name="tftp-client" accept']}
    assert 'tftp-client' in get_used_services(conf, True)


def test_get_used_services_tuple_zone():
    conf = (None, None, None, None, None,
            ['tftp-client'],  # conf[5], services
            None, None, None, None,
            ['dummy0'],  # conf[10], interfaces
            [],  # conf[11], sources
            [],  # conf[12], rules_str
            None, None, None)
    assert 'tftp-client' in get_used_services_tuple(conf, True)

    conf = (None, None, None, None, None,
            [],  # conf[5], services
            None, None, None, None,
            [],  # conf[10], interfaces
            ['10.1.2.0/24'],  # conf[11], sources
            ['rule family="ipv4" source address="10.1.1.0/24" service name="tftp-client" reject'],
            None, None, None)
    assert 'tftp-client' in get_used_services_tuple(conf, True)

    conf = (None, None, None, None, None,
            [],  # conf[5], services
            None, None, None, None,
            ['dummy0'],  # conf[10], interfaces
            ['fd00::/8'],  # conf[11], sources
            ['rule service name="ssh" accept', 'rule service name="tftp-client" accept'],  # conf[12], rules_str
            None, None, None)
    assert 'tftp-client' in get_used_services_tuple(conf, True)


def test_get_used_services_zone_negative():
    conf = {'interfaces': ['dummy0'],
            'services': ['https']}
    assert 'tftp-client' not in get_used_services(conf, True)

    conf = {'sources': ['10.1.2.0/24'],
            'rules_str': ['rule family="ipv4" source address="10.1.1.0/24" service name="ssh" reject'],
            'services': ['https']}
    assert 'tftp-client' not in get_used_services(conf, True)

    conf = {'interfaces': ['dummy0'],
            'sources': ['fd00::/8'],
            'rules_str': ['rule service name="ssh" accept',
                          'rule service name="http" accept']}
    assert 'tftp-client' not in get_used_services(conf, True)


def test_get_used_services_tuple_zone_negative():
    conf = (None, None, None, None, None,
            ['https'],  # conf[5], services
            None, None, None, None,
            ['dummy0'],  # conf[10], interfaces
            [],  # conf[11], sources
            [],  # conf[12], rules_str
            None, None, None)
    assert 'tftp-client' not in get_used_services_tuple(conf, True)

    conf = {'sources': ['10.1.2.0/24'],
            'rules_str': ['rule family="ipv4" source address="10.1.1.0/24" service name="ssh" reject'],
            'services': ['https']}
    conf = (None, None, None, None, None,
            ['https'],  # conf[5], services
            None, None, None, None,
            [],  # conf[10], interfaces
            ['10.1.2.0/24'],  # conf[11], sources
            ['rule family="ipv4" source address="10.1.1.0/24" service name="ssh" reject'],  # conf[12], rules_str
            None, None, None)
    assert 'tftp-client' not in get_used_services_tuple(conf, True)

    conf = (None, None, None, None, None,
            [],  # conf[5], services
            None, None, None, None,
            ['dummy0'],  # conf[10], interfaces
            ['fd00::/8'],  # conf[11], sources
            ['rule service name="ssh" accept', 'rule service name="http" accept'],  # conf[12], rules_str
            None, None, None)
    assert 'tftp-client' not in get_used_services_tuple(conf, True)


def test_get_used_services_policy():
    conf = {'services': ['tftp-client']}
    assert 'tftp-client' in get_used_services(conf, False)

    conf = {'rich_rules': ['rule family="ipv4" source address="10.1.1.0/24" service name="tftp-client" reject']}
    assert 'tftp-client' in get_used_services(conf, False)

    conf = {'rich_rules': ['rule service name="ssh" accept',
                           'rule service name="tftp-client" accept']}
    assert 'tftp-client' in get_used_services(conf, False)


def test_get_used_services_policy_negative():
    conf = {}
    assert 'tftp-client' not in get_used_services(conf, False)

    conf = {'services': []}
    assert 'tftp-client' not in get_used_services(conf, False)

    conf = {'services': ['ssh']}
    assert 'tftp-client' not in get_used_services(conf, False)

    conf = {'rich_rules': ['rule family="ipv4" source address="10.1.1.0/24" service name="http" reject']}
    assert 'tftp-client' not in get_used_services(conf, False)

    conf = {'rich_rules': ['rule service name="ssh" accept',
                           'rule service name="https" accept']}
    assert 'tftp-client' not in get_used_services(conf, False)



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class NtpMigrationDecision(Model):
    topic = SystemInfoTopic
    migrate_services = fields.List(fields.String())
    config_tgz64 = fields.String()



import pytest
from six import text_type

from leapp.models import BrlttyMigrationDecision


@pytest.mark.parametrize('test_input,expected,migrate_bt,migrate_espeak', [
    ('braille-device bth:AB-cd:ef:01:23:45\n', 'braille-device bluetooth:AB-cd:ef:01:23:45', True, False),
    ('braille-device bluez:AB-cd:ef:01:23:45\n', 'braille-device bluetooth:AB-cd:ef:01:23:45', True, False),
    ('speech-driver es\n', 'speech-driver en', False, True),
    ('braille-device bth:AB-cd:ef:01:23:45\nbraille-device bluez:AB-cd:ef:01:23:45\nspeech-driver es\n',
        'braille-device bluetooth:AB-cd:ef:01:23:45\nbraille-device bluetooth:AB-cd:ef:01:23:45\nspeech-driver es',
        True, False),
    ('braille-device bth:AB-cd:ef:01:23:45\nbraille-device bluez:AB-cd:ef:01:23:45\nspeech-driver es\n',
        'braille-device bth:AB-cd:ef:01:23:45\nbraille-device bluez:AB-cd:ef:01:23:45\nspeech-driver en', False, True),
    ('braille-device bth:AB-cd:ef:01:23:45\nbraille-device bluez:AB-cd:ef:01:23:45\nspeech-driver es\n',
        'braille-device bluetooth:AB-cd:ef:01:23:45\nbraille-device bluetooth:AB-cd:ef:01:23:45\nspeech-driver en',
        True, True),
])
def test_actor_migrate(tmpdir, test_input, expected, migrate_bt, migrate_espeak, current_actor_context):
    brltty_conf = text_type(tmpdir.join('brltty.conf'))
    with open(brltty_conf, 'w') as file_out:
        file_out.write(test_input)
    current_actor_context.feed(BrlttyMigrationDecision(migrate_file=brltty_conf, migrate_bt=migrate_bt,
                               migrate_espeak=migrate_espeak))
    current_actor_context.run()
    with open(brltty_conf, 'r') as file_in:
        data = file_in.read().strip()
    assert expected == data



from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import DNFWorkaround, InstalledRPM

# maps target version to keys obsoleted in that version
OBSOLETED_KEYS_MAP = {
    7: [],
    8: [
        "gpg-pubkey-2fa658e0-45700c69",
        "gpg-pubkey-37017186-45761324",
        "gpg-pubkey-db42a60e-37ea5438",
    ],
    9: ["gpg-pubkey-d4082792-5b32db75"],
}


def _get_obsolete_keys():
    """
    Return keys obsoleted in target and previous versions
    """
    keys = []
    for version in range(7, int(get_target_major_version()) + 1):
        for key in OBSOLETED_KEYS_MAP[version]:
            name, version, release = key.rsplit("-", 2)
            if has_package(InstalledRPM, name, version=version, release=release):
                keys.append(key)

    return keys


def register_dnfworkaround(keys):
    api.produce(
        DNFWorkaround(
            display_name="remove obsolete RPM GPG keys from RPM DB",
            script_path=api.current_actor().get_common_tool_path("removerpmgpgkeys"),
            script_args=keys,
        )
    )


def process():
    keys = _get_obsolete_keys()
    if not keys:
        return

    register_dnfworkaround(keys)



import os

import pytest

from leapp.libraries.actor import setetcreleasever
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, logger_mocked
from leapp.libraries.stdlib import api
from leapp.models import (
    PkgManagerInfo,
    RHUIInfo,
    TargetRHUIPostInstallTasks,
    TargetRHUIPreInstallTasks,
    TargetRHUISetupInfo
)

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


class mocked_set_releasever(object):
    def __init__(self):
        self.content = None

    def __call__(self, content):
        self.content = content


# def test_set_releasever(monkeypatch):

#     target = '8.0'
#     expected_rel_ver = '8.0'
#     monkeypatch.setattr(setetcreleasever, '_set_releasever', mocked_set_releasever())

#     setetcreleasever.process(target)
#     assert expected_rel_ver == setetcreleasever._set_releasever.content


def test_set_releasever(monkeypatch, current_actor_context):
    preinstall_tasks = TargetRHUIPreInstallTasks()
    postinstall_tasks = TargetRHUIPostInstallTasks()
    setup_info = TargetRHUISetupInfo(preinstall_tasks=preinstall_tasks, postinstall_tasks=postinstall_tasks)
    rhui_info = RHUIInfo(provider='aws',
                         src_client_pkg_names=['rh-amazon-rhui-client'],
                         target_client_pkg_names=['rh-amazon-rhui-client'],
                         target_client_setup_info=setup_info)

    msgs = [rhui_info, PkgManagerInfo(etc_releasever='7.7')]

    expected_rel_ver = '8.0'
    monkeypatch.setattr(setetcreleasever, '_set_releasever', mocked_set_releasever())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        msgs=msgs, dst_ver=expected_rel_ver
        )
    )
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    setetcreleasever.process()

    assert expected_rel_ver == setetcreleasever._set_releasever.content
    assert not api.current_logger.dbgmsg


def test_no_set_releasever(monkeypatch, current_actor_context):

    expected_rel_ver = '8.0'
    monkeypatch.setattr(setetcreleasever, '_set_releasever', mocked_set_releasever())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(dst_ver=expected_rel_ver))
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    setetcreleasever.process()

    assert not setetcreleasever._set_releasever.content
    assert api.current_logger.dbgmsg



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class BlackListCA(Model):
    """
    Provides an entry for all disabled CAs in one of blacklist directoriesi
    which needs to be moved to blocklist.
    """
    topic = SystemInfoTopic

    source = fields.String()
    """
    The full path to the file in the blacklist directory.
    """

    sourceDir = fields.String()
    """
    The path of the blacklist directory where source resides.
    """

    target = fields.String()
    """
    The full path to where the file should be migrated to.
    """

    targetDir = fields.String()
    """
    The path of the blocklist directory where the target resides
    """



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class UsedRepository(Model):
    """
    Describe list of current packages installed from a specific repository
    """

    topic = SystemInfoTopic

    repository = fields.String()
    packages = fields.List(fields.String(), default=[])


class UsedRepositories(Model):
    """
    Describe list of used repositories in the current system
    """

    topic = SystemInfoTopic

    repositories = fields.List(fields.Model(UsedRepository), default=[])



# -*- coding: UTF-8 -*-
import os.path
import sys

from leapp.libraries.common.utils import config, logging_handler
from leapp.libraries.stdlib.call import STDERR, STDOUT


def test_logging_handler(capfdbinary, monkeypatch):
    debug_on = False
    monkeypatch.setattr(config, 'is_debug', lambda: debug_on)
    panagrams_path = os.path.join(os.path.dirname(__file__), 'panagrams')
    with open(panagrams_path, 'rb') as f:
        test_data = f.read()

    bin_test_data = test_data
    if isinstance(test_data, str) and str is not bytes:
        bin_test_data = test_data.encode()
    # Should not log anything
    for c in range(0, len(bin_test_data)):
        logging_handler((None, STDERR), bin_test_data[c:c+1])
    debug_on = True
    for c in range(0, len(bin_test_data)):
        logging_handler((None, STDERR), bin_test_data[c:c+1])
    for c in range(0, len(bin_test_data)):
        logging_handler((None, STDOUT), bin_test_data[c:c+1])
    captured = capfdbinary.readouterr()
    assert captured.out == bin_test_data
    assert captured.err == bin_test_data


def test_logging_handler_arrow(capfdbinary, monkeypatch):
    debug_on = False
    monkeypatch.setattr(config, 'is_debug', lambda: debug_on)
    # → => '\xe2\x86\x92'
    if sys.version_info > (3, 0):
        buf = b'\xe2\x86\x92'
    else:
        buf = '\xe2\x86\x92'

    for i in range(0, len(buf)):
        debug_on = False
        logging_handler((None, STDERR), buf[i:i+1])
        debug_on = True
        logging_handler((None, STDERR), buf[i:i+1])
        logging_handler((None, STDOUT), buf[i:i+1])
    captured = capfdbinary.readouterr()
    assert captured.err == buf
    assert captured.out == buf



from leapp.libraries.stdlib import run


def update_kernel_config(path):
    """ Update DEFAULTKERNEL entry at provided config file """
    run(['/bin/sed',
         '-i',
         's/^DEFAULTKERNEL=kernel$/DEFAULTKERNEL=kernel-core/g',
         path])



from leapp import reporting
from leapp.libraries.actor import checkchrony
from leapp.libraries.common.testutils import create_report_mocked


def test_uninstalled(monkeypatch):
    for config_default in (False, True):
        monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
        monkeypatch.setattr(checkchrony, 'is_config_default', lambda: config_default)

        checkchrony.check_chrony(False)

        assert reporting.create_report.called == 0


def test_installed_defconf(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(checkchrony, 'is_config_default', lambda: True)

    checkchrony.check_chrony(True)

    assert reporting.create_report.called == 1
    assert reporting.create_report.report_fields['title'] == 'chrony using default configuration'


def test_installed_nodefconf(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(checkchrony, 'is_config_default', lambda: False)

    checkchrony.check_chrony(True)

    assert reporting.create_report.called == 1
    assert reporting.create_report.report_fields['title'] == 'chrony using non-default configuration'



import contextlib
import os
import shutil
from collections import namedtuple

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import mounting, utils
from leapp.libraries.common.config import get_env
from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.stdlib import api, CalledProcessError, run

OVERLAY_DO_NOT_MOUNT = ('tmpfs', 'devtmpfs', 'devpts', 'sysfs', 'proc', 'cramfs', 'sysv', 'vfat')

# NOTE(pstodulk): what about using more closer values and than just multiply
# the final result by magical constant?... this number is most likely going to
# be lowered and affected by XFS vs EXT4 FSs that needs different spaces each
# of them.
_MAGICAL_CONSTANT_OVL_SIZE = 128
"""
Average size of created disk space images.

The size can be lower or higher - usually lower. The value is higher as we want
to rather prevent future actions in advance instead of resolving later issues
with the missing space.

It's possible that in future we implement better heuristic that will guess
the needed space based on size of each FS. I have been thinking to lower
the value, as in my case most of partitions where we do not need to do
write operations consume just ~ 33MB. However, I decided to keep it as it is
for now to stay on the safe side.
"""

_MAGICAL_CONSTANT_MIN_CONTAINER_SIZE_8 = 3200
"""
Average space consumed to create target el8userspace container installation + pkg downloads.

Minimal container size is approx. 1GiB without download of packages for the upgrade
(and without pkgs for the initramfs creation). The total size of the container
  * with all pkgs downloaded
  * final initramfs installed package set
  * created the upgrade initramfs
is for the minimal system
  * ~ 2.9 GiB for IPU 7 -> 8
  * ~ 1.8 GiB for IPU 8 -> 9
when no other extra packages are installed for the needs of the upgrade.
Keeping in mind that during the upgrade initramfs creation another 400+ MiB
is consumed temporarily.

Using higher value to cover also the space that consumes leapp.db records.

This constant is really magical and the value can be changed in future.
"""

_MAGICAL_CONSTANT_MIN_CONTAINER_SIZE_9 = 2200
"""
Average space consumed to create target el9userspace container installation + pkg downloads.

See _MAGICAL_CONSTANT_MIN_CONTAINER_SIZE_8 for more details.
"""

_MAGICAL_CONSTANT_MIN_PROTECTED_SIZE = 200
"""
This is the minimal size (in MiB) that will be always reserved for /var/lib/leapp

In case the size of the container is larger than _MAGICAL_CONSTANT_MIN_PROTECTED_SIZE
or close to that size, stay always with this minimal protected size defined by
this constant.
"""


MountPoints = namedtuple('MountPoints', ['fs_file', 'fs_vfstype'])


def _get_min_container_size():
    if get_target_major_version() == '8':
        return _MAGICAL_CONSTANT_MIN_CONTAINER_SIZE_8
    return _MAGICAL_CONSTANT_MIN_CONTAINER_SIZE_9


def get_recommended_leapp_free_space(userspace_path=None):
    """
    Return recommended free space for the target container (+ pkg downloads)

    If the path to the container is set, the returned value is updated to
    reflect already consumed space by the installed container. In case the
    container is bigger than the minimal protected size, return at least
    `_MAGICAL_CONSTANT_MIN_PROTECTED_SIZE`.

    It's not recommended to use this function except official actors managed
    by OAMG group in github.com/oamg/leapp-repository. This function can be
    changed in future, ignoring the deprecation process.

    TODO(pstodulk): this is so far the best trade off between stay safe and do
    do not consume too much space. But need to figure out cost of the time
    consumption.

    TODO(pstodulk): check we are not negatively affected in case of downloaded
    rpms. We want to prevent situations when we say that customer has enough
    space for the first run and after the download of packages we inform them
    they do not have enough free space anymore. Note: such situation can be
    valid in specific cases - e.g. the space is really consumed already e.g. by
    leapp.db that has been executed manytimes.

    :param userspace_path: Path to the userspace container.
    :type userspace_path: str
    :rtype: int
    """
    min_cont_size = _get_min_container_size()
    if not userspace_path or not os.path.exists(userspace_path):
        return min_cont_size
    try:
        # ignore symlinks and other partitions to be sure we calculate the space
        # in reasonable time
        cont_size = run(['du', '-sPmx', userspace_path])['stdout'].split()[0]
        # the obtained number is in KiB. But we want to work with MiBs rather.
        cont_size = int(cont_size)
    except (OSError, CalledProcessError):
        # do not care about failed cmd, in such a case, just act like userspace_path
        # has not been set
        api.current_logger().warning(
            'Cannot calculate current container size to estimate correctly required space.'
            ' Working with the default: {} MiB'
            .format(min_cont_size)
        )
        return min_cont_size
    if cont_size < 0:
        api.current_logger().warning(
            'Cannot calculate the container size - negative size obtained: {}.'
            ' Estimate the required size based on the default value: {} MiB'
            .format(cont_size, min_cont_size)
        )
        return min_cont_size
    prot_size = min_cont_size - cont_size
    if prot_size < _MAGICAL_CONSTANT_MIN_PROTECTED_SIZE:
        api.current_logger().debug(
            'The size of the container is higher than the expected default.'
            ' Use the minimal protected size instead: {} MiB.'
            .format(_MAGICAL_CONSTANT_MIN_PROTECTED_SIZE)
        )
        return _MAGICAL_CONSTANT_MIN_PROTECTED_SIZE
    return prot_size


def _get_fspace(path, convert_to_mibs=False, coefficient=1):
    """
    Return the free disk space on given path.

    The default is in bytes, but if convert_to_mibs is True, return MiBs instead.

    Raises OSError if nothing exists on the given `path`.

    :param path: Path to an existing file or directory
    :type path: str
    :param convert_to_mibs: If True, convert the value to MiBs
    :type convert_to_mibs: bool
    :param coefficient: Coefficient to multiply the free space (e.g. 0.9 to have it 10% lower). Max: 1
    :type coefficient: float
    :rtype: int
    """
    stat = os.statvfs(path)

    # TODO(pstodulk): discuss the function params
    coefficient = min(coefficient, 1)
    fspace_bytes = int(stat.f_frsize * stat.f_bavail * coefficient)
    if convert_to_mibs:
        return int(fspace_bytes / 1024 / 1024)  # noqa: W1619; pylint: disable=old-division
    return fspace_bytes


def _ensure_enough_diskimage_space(space_needed, directory):
    # TODO(pstodulk): update the error msg/details
    # imagine situation we inform user we need at least 800MB,
    # so they clean /var/lib/leapp/* which can provide additional space,
    # but the calculated required free space takes the existing content under
    # /var/lib/leapp/ into account, so the next error msg could say:
    #    needed at least 3400 MiB - which could be confusing for users.
    if _get_fspace(directory) < (space_needed * 1024 * 1024):
        message = (
            'Not enough space available on {directory}: Needed at least {space_needed} MiB.'
            .format(directory=directory, space_needed=space_needed)
        )
        details = {'detail': (
            'The file system hosting the {directory} directory does not contain'
            ' enough free space to proceed all parts of the in-place upgrade.'
            ' Note the calculated required free space is the minimum derived'
            ' from upgrades of minimal systems and the actual needed free'
            ' space could be higher.'
            '\nNeeded at least: {space_needed} MiB.'
            '\nSuggested free space: {suggested} MiB (or more).'
            .format(space_needed=space_needed, directory=directory, suggested=space_needed + 1000)
        )}
        if get_env('LEAPP_OVL_SIZE', None):
            # LEAPP_OVL_SIZE has not effect as we use sparse files now.
            details['note'] = 'The LEAPP_OVL_SIZE environment variable has no effect anymore.'
        api.current_logger().error(message)
        raise StopActorExecutionError(message, details=details)


def _get_mountpoints(storage_info):
    mount_points = set()
    for entry in storage_info.fstab:
        if os.path.isdir(entry.fs_file) and entry.fs_vfstype not in OVERLAY_DO_NOT_MOUNT:
            mount_points.add(MountPoints(entry.fs_file, entry.fs_vfstype))
        elif os.path.isdir(entry.fs_file) and entry.fs_vfstype == 'vfat':
            # VFAT FS is not supported to be used for any system partition,
            # so we can safely ignore it
            api.current_logger().warning(
                'Ignoring vfat {} filesystem mount during upgrade process'.format(entry.fs_file)
            )

    return list(mount_points)


def _mount_name(mountpoint):
    return 'root{}'.format(mountpoint.replace('/', '_'))


def _mount_dir(mounts_dir, mountpoint):
    return os.path.join(mounts_dir, _mount_name(mountpoint))


def _get_scratch_mountpoint(mount_points, dir_path):
    for mp in sorted(mount_points, reverse=True):
        # we are sure that mountpoint != dir_path in this case, as the latest
        # valid mountpoint customers could create is the parent directory
        mod_mp = mp if mp[-1] == '/' else '{}/'.format(mp)
        if dir_path.startswith(mod_mp):
            # longest first, so the first one we find, is the last mp on the path
            return mp
    return None  # making pylint happy; this is basically dead code


def _prepare_required_mounts(scratch_dir, mounts_dir, storage_info, scratch_reserve):
    """
    Create disk images and loop mount them.

    Ensure to create disk image for each important mountpoint configured
    in fstab (excluding fs types noted in `OVERLAY_DO_NOT_MOUNT`).
    Disk images reflect the free space of related partition/volume. In case
    of partition hosting /var/lib/leapp/* calculate the free space value
    taking `scratch_reserve` into account, as during the run of the tooling,
    we will be consuming the space on the partition and we want to be more
    sure that we do not consume all the space on the partition during the
    execution - so we reduce the risk we affect run of other applications
    due to missing space.

    Note: the partition hosting the scratch dir is expected to be the same
    partition that is hosting the target userspace container, but it does not
    have to be true if the code changes. Right now, let's live with that.

    See `_create_mount_disk_image` docstring for additional more details.

    :param scratch_dir: Path to the scratch directory.
    :type scratch_dir: str
    :param mounts_dir: Path to the directory supposed to be a mountpoint.
    :type mounts_dir: str
    :param storage_info: The StorageInfo message.
    :type storage_info: leapp.models.StorageInfo
    :param scratch_reserve: Number of MB that should be extra reserved in a partition hosting the scratch_dir.
    :type scratch_reserve: Optional[int]
    """
    mount_points = sorted([mp.fs_file for mp in _get_mountpoints(storage_info)])
    scratch_mp = _get_scratch_mountpoint(mount_points, scratch_dir)
    disk_images_directory = os.path.join(scratch_dir, 'diskimages')

    # Ensure we cleanup old disk images before we check for space constraints.
    # NOTE(pstodulk): Could we improve the process so we create imgs & calculate
    # the required disk space just once during each leapp (pre)upgrade run?
    run(['rm', '-rf', disk_images_directory])
    _create_diskimages_dir(scratch_dir, disk_images_directory)

    # TODO(pstodulk): update the calculation for bind mounted mount_points (skip)
    # basic check whether we have enough space at all
    space_needed = scratch_reserve + _MAGICAL_CONSTANT_OVL_SIZE * len(mount_points)
    _ensure_enough_diskimage_space(space_needed, scratch_dir)

    # free space required on this partition should not be affected by durin the
    # upgrade transaction execution by space consumed on creation of disk images
    # as disk images are cleaned in the end of this functions,
    # but we want to reserve some space in advance.
    scratch_disk_size = _get_fspace(scratch_dir, convert_to_mibs=True) - scratch_reserve

    result = {}
    for mountpoint in mount_points:
        # keep the info about the free space rather 5% lower than the real value
        disk_size = _get_fspace(mountpoint, convert_to_mibs=True, coefficient=0.95)
        if mountpoint == scratch_mp:
            disk_size = scratch_disk_size
        image = _create_mount_disk_image(disk_images_directory, mountpoint, disk_size)
        result[mountpoint] = mounting.LoopMount(
            source=image,
            target=_mount_dir(mounts_dir, mountpoint)
        )
    return result


@contextlib.contextmanager
def _build_overlay_mount(root_mount, mounts):
    if not root_mount:
        raise StopActorExecutionError('Root mount point has not been prepared for overlayfs.')
    if not mounts:
        yield root_mount
    else:
        current = list(mounts.keys())[0]
        current_mount = mounts.pop(current)
        name = _mount_name(current)
        with current_mount:
            with mounting.OverlayMount(name=name, source=current, workdir=current_mount.target) as overlay:
                with mounting.BindMount(source=overlay.target,
                                        target=os.path.join(root_mount.target, current.lstrip('/'))):
                    with _build_overlay_mount(root_mount, mounts) as mount:
                        yield mount


def cleanup_scratch(scratch_dir, mounts_dir):
    """
    Function to cleanup the scratch directory

    If the mounts_dir is a mountpoint, unmount it first.

    :param scratch_dir: Path to the scratch directory.
    :type scratch_dir: str
    :param mounts_dir: Path to the directory supposed to be a mountpoint.
    :type mounts_dir: str
    """
    api.current_logger().debug('Cleaning up mounts')
    if os.path.ismount(mounts_dir):
        # TODO(pstodulk): this is actually obsoleted for years. mounts dir
        # is not mountpoit anymore, it contains mountpoints. But in time of
        # this call all MPs should be already umounted as the solution has been
        # changed also (all MPs are handled by context managers). This code
        # is basically dead, so keeping it as it does not hurt us now.
        api.current_logger().debug('Mounts directory is a mounted disk image - Unmounting.')
        try:
            run(['/bin/umount', '-fl', mounts_dir])
            api.current_logger().debug('Unmounted mounted disk image.')
        except (OSError, CalledProcessError) as e:
            api.current_logger().warning('Failed to umount %s - message: %s', mounts_dir, str(e))
    if get_env('LEAPP_DEVEL_KEEP_DISK_IMGS', None) == '1':
        # NOTE(pstodulk): From time to time, it helps me with some experiments
        return
    api.current_logger().debug('Recursively removing scratch directory %s.', scratch_dir)
    shutil.rmtree(scratch_dir, onerror=utils.report_and_ignore_shutil_rmtree_error)
    api.current_logger().debug('Recursively removed scratch directory %s.', scratch_dir)


def _format_disk_image_ext4(diskimage_path):
    """
    Format the specified disk image with Ext4 filesystem.

    The formatted file system is optimized for operations we want to do and
    mainly for the space it needs to take for the initialisation. So use 32MiB
    journal (that's enough for us as we do not plan to do too many operations
    inside) for any size of the disk image. Also the lazy
    initialisation is disabled. The formatting will be slower, but it helps
    us to estimate better the needed amount of the space for other actions
    done later.
    """
    api.current_logger().debug('Creating ext4 filesystem in disk image at %s', diskimage_path)
    cmd = [
        '/sbin/mkfs.ext4',
        '-J', 'size=32',
        '-E', 'lazy_itable_init=0,lazy_journal_init=0',
        '-F', diskimage_path
    ]
    try:
        utils.call_with_oserror_handled(cmd=cmd)
    except CalledProcessError as e:
        # FIXME(pstodulk): taken from original, but %s seems to me invalid here
        api.current_logger().error('Failed to create ext4 filesystem in %s', diskimage_path, exc_info=True)
        raise StopActorExecutionError(
            message='Cannot create Ext4 filesystem in {}'.format(diskimage_path),
            details={
                'error message': str(e),
            }
        )


def _format_disk_image_xfs(diskimage_path):
    """
    Format the specified disk image with XFS filesystem.

    Set journal just to 32MiB always as we will not need to do too many operation
    inside, so 32MiB should enough for us.
    """
    api.current_logger().debug('Creating XFS filesystem in disk image at %s', diskimage_path)
    cmd = ['/sbin/mkfs.xfs', '-l', 'size=32m', '-f', diskimage_path]
    try:
        utils.call_with_oserror_handled(cmd=cmd)
    except CalledProcessError as e:
        # FIXME(pstodulk): taken from original, but %s seems to me invalid here
        api.current_logger().error('Failed to create XFS filesystem %s', diskimage_path, exc_info=True)
        raise StopActorExecutionError(
            message='Cannot create XFS filesystem in {}'.format(diskimage_path),
            details={
                'error message': str(e),
            }
        )


def _create_mount_disk_image(disk_images_directory, path, disk_size):
    """
    Creates the mount disk image and return path to it.

    The disk image is represented by a sparse file which apparent size
    corresponds usually to the free space of a particular partition/volume it
    represents - in this function it's set by `disk_size` parameter, which should
    be int representing the free space in MiBs.

    The created disk image is formatted with XFS (default) or Ext4 FS
    and it's supposed to be used for write directories of an overlayfs built
    above it.

    If the disk_size is lower than 130 MiBs, the disk size is automatically
    set to 130 MiBs to be able to format it correctly.

    The disk image is formatted with Ext4 if (envar) `LEAPP_OVL_IMG_FS_EXT4=1`.

    :param disk_images_directory: Path to the directory where disk images should be stored.
    :type disk_images_directory: str
    :param path: Path to the mountpoint of the original (host/source) partition/volume
    :type path: str
    :param disk_size: Apparent size of the disk img in MiBs
    :type disk_size: int
    :return: Path to the created disk image
    :rtype: str
    """
    if disk_size < 130:
        # NOTE(pstodulk): SEATBELT
        # min. required size for current params to format a disk img with a FS:
        #   XFS  -> 130 MiB
        #   EXT4 -> 70  MiB
        # so let's stick to 130 always. This is expected to happen when:
        #  * the free space on a system mountpoint is really super small, but if
        #    such a mounpoint contains a content installed by packages, most
        #    likely the msg about not enough free space is raised
        #  * the mountpoint is actually no important at all, could be possibly
        #    read only (e.g. ISO), or it's an FS type that should be covered by
        #    OVERLAY_DO_NOT_MOUNT
        #  * most common case important for us here could be /boot, but that's
        #    covered already in different actors/checks, so it should not be
        #    problem either
        # NOTE(pstodulk): In case the formatting params are modified,
        # the minimal required size could be different
        api.current_logger().warning(
            'The apparent size for the disk image representing {path}'
            ' is too small ({disk_size} MiBs) for a formatting. Setting 130 MiBs instead.'
            .format(path=path, disk_size=disk_size)
        )
        disk_size = 130
    diskimage_path = os.path.join(disk_images_directory, _mount_name(path))
    cmd = [
        '/bin/dd',
        'if=/dev/zero', 'of={}'.format(diskimage_path),
        'bs=1M', 'count=0', 'seek={}'.format(disk_size)
    ]
    hint = (
        'Please ensure that there is enough diskspace on the partition hosting'
        'the {} directory.'
        .format(disk_images_directory)
    )

    api.current_logger().debug('Attempting to create disk image at %s', diskimage_path)
    utils.call_with_failure_hint(cmd=cmd, hint=hint)

    if get_env('LEAPP_OVL_IMG_FS_EXT4', '0') == '1':
        # This is alternative to XFS in case we find some issues, to be able
        # to switch simply to Ext4, so we will be able to simple investigate
        # possible issues between overlay <-> XFS if any happens.
        _format_disk_image_ext4(diskimage_path)
    else:
        _format_disk_image_xfs(diskimage_path)

    return diskimage_path


def _create_diskimages_dir(scratch_dir, diskimages_dir):
    """
    Prepares directories for disk images
    """
    api.current_logger().debug('Creating disk images directory.')
    try:
        utils.makedirs(diskimages_dir)
        api.current_logger().debug('Done creating disk images directory.')
    except OSError:
        api.current_logger().error('Failed to create disk images directory %s', diskimages_dir, exc_info=True)

        # This is an attempt for giving the user a chance to resolve it on their own
        raise StopActorExecutionError(
            message='Failed to prepare environment for package download while creating directories.',
            details={
                'hint': 'Please ensure that {scratch_dir} is empty and modifiable.'.format(scratch_dir=scratch_dir)
            }
        )


def _create_mounts_dir(scratch_dir, mounts_dir):
    """
    Prepares directories for mounts
    """
    api.current_logger().debug('Creating mount directories.')
    try:
        utils.makedirs(mounts_dir)
        api.current_logger().debug('Done creating mount directories.')
    except OSError:
        api.current_logger().error('Failed to create mounting directories %s', mounts_dir, exc_info=True)

        # This is an attempt for giving the user a chance to resolve it on their own
        raise StopActorExecutionError(
            message='Failed to prepare environment for package download while creating directories.',
            details={
                'hint': 'Please ensure that {scratch_dir} is empty and modifiable.'.format(scratch_dir=scratch_dir)
            }
        )


@contextlib.contextmanager
def _mount_dnf_cache(overlay_target):
    """
    Convenience context manager to ensure bind mounted /var/cache/dnf and removal of the mount.
    """
    with mounting.BindMount(
            source='/var/cache/dnf',
            target=os.path.join(overlay_target, 'var', 'cache', 'dnf')) as cache_mount:
        yield cache_mount


@contextlib.contextmanager
def create_source_overlay(mounts_dir, scratch_dir, xfs_info, storage_info, mount_target=None, scratch_reserve=0):
    """
    Context manager that prepares the source system overlay and yields the mount.

    The in-place upgrade itself requires to do some changes on the system to be
    able to perform the in-place upgrade itself - or even to be able to evaluate
    if the system is possible to upgrade. However, we do not want to (and must not)
    change the original system until we pass beyond the point of not return.

    For that purposes we have to create a layer above the real host file system,
    where we can safely perform all operations without affecting the system
    setup, rpm database, etc. Currently overlay (OVL) technology showed it is
    capable to handle our requirements good enough - with some limitations.

    This function prepares a disk image and an overlay layer for each
    mountpoint configured in /etc/fstab, excluding those with FS type noted
    in the OVERLAY_DO_NOT_MOUNT set. Such prepared OVL images are then composed
    together to reflect the real host filesystem. In the end everything is cleaned.

    The new solution can be now problematic for system with too many partitions
    and loop devices. For such systems we keep for now the possibility of the
    fallback to an old solution, which has however number of issues that are
    fixed by the new design. To fallback to the old solution, set envar:
        LEAPP_OVL_LEGACY=1

    Disk images created for OVL are formatted with XFS by default. In case of
    problems, it's possible to switch to Ext4 FS using:
        LEAPP_OVL_IMG_FS_EXT4=1

    :param mounts_dir: Absolute path to the directory under which all mounts should happen.
    :type mounts_dir: str
    :param scratch_dir: Absolute path to the directory in which all disk and OVL images are stored.
    :type scratch_dir: str
    :param xfs_info: The XFSPresence message.
    :type xfs_info: leapp.models.XFSPresence
    :param storage_info: The StorageInfo message.
    :type storage_info: leapp.models.StorageInfo
    :param mount_target: Directory to which whole source OVL layer should be bind mounted.
                         If None (default), mounting.NullMount is created instead
    :type mount_target: Optional[str]
    :param scratch_reserve: Number of MB that should be extra reserved in a partition hosting the scratch_dir.
    :type scratch_reserve: Optional[int]
    :rtype: mounting.BindMount or mounting.NullMount
    """
    api.current_logger().debug('Creating source overlay in {scratch_dir} with mounts in {mounts_dir}'.format(
        scratch_dir=scratch_dir, mounts_dir=mounts_dir))
    try:
        _create_mounts_dir(scratch_dir, mounts_dir)
        if get_env('LEAPP_OVL_LEGACY', '0') != '1':
            mounts = _prepare_required_mounts(scratch_dir, mounts_dir, storage_info, scratch_reserve)
        else:
            # fallback to the deprecated OVL solution
            mounts = _prepare_required_mounts_old(scratch_dir, mounts_dir, _get_mountpoints(storage_info), xfs_info)
        with mounts.pop('/') as root_mount:
            with mounting.OverlayMount(name='system_overlay', source='/', workdir=root_mount.target) as root_overlay:
                if mount_target:
                    target = mounting.BindMount(source=root_overlay.target, target=mount_target)
                else:
                    target = mounting.NullMount(target=root_overlay.target)
                with target:
                    with _build_overlay_mount(root_overlay, mounts) as overlay:
                        with _mount_dnf_cache(overlay.target):
                            yield overlay
    except Exception:
        cleanup_scratch(scratch_dir, mounts_dir)
        raise
    # cleanup always now
    cleanup_scratch(scratch_dir, mounts_dir)


# #############################################################################
# Deprecated OVL solution ...
# This is going to be removed in future as the whole functionality is going to
# be replaced by new one. The problem is that the new solution can potentially
# negatively affect systems with many loop mountpoints, so let's keep this
# as a workaround for now. I am separating the old and new code in this way
# to make the future removal easy.
# The code below is triggered when LEAPP_OVL_LEGACY=1 envar is set.
# IMPORTANT: Before an update of functions above, ensure the functionality of
# the code below is not affected, otherwise copy the function below with the
# "_old" suffix.
# #############################################################################
def _ensure_enough_diskimage_space_old(space_needed, directory):
    stat = os.statvfs(directory)
    if (stat.f_frsize * stat.f_bavail) < (space_needed * 1024 * 1024):
        message = ('Not enough space available for creating required disk images in {directory}. ' +
                   'Needed: {space_needed} MiB').format(space_needed=space_needed, directory=directory)
        api.current_logger().error(message)
        raise StopActorExecutionError(message)


def _overlay_disk_size_old():
    """
    Convenient function to retrieve the overlay disk size
    """
    try:
        env_size = get_env('LEAPP_OVL_SIZE', '2048')
        disk_size = int(env_size)
    except ValueError:
        disk_size = 2048
        api.current_logger().warning(
            'Invalid "LEAPP_OVL_SIZE" environment variable "%s". Setting default "%d" value', env_size, disk_size
        )
    return disk_size


def _create_diskimages_dir_old(scratch_dir, diskimages_dir):
    """
    Prepares directories for disk images
    """
    api.current_logger().debug('Creating disk images directory.')
    try:
        utils.makedirs(diskimages_dir)
        api.current_logger().debug('Done creating disk images directory.')
    except OSError:
        api.current_logger().error('Failed to create disk images directory %s', diskimages_dir, exc_info=True)

        # This is an attempt for giving the user a chance to resolve it on their own
        raise StopActorExecutionError(
            message='Failed to prepare environment for package download while creating directories.',
            details={
                'hint': 'Please ensure that {scratch_dir} is empty and modifiable.'.format(scratch_dir=scratch_dir)
            }
        )


def _create_mount_disk_image_old(disk_images_directory, path):
    """
    Creates the mount disk image, for cases when we hit XFS with ftype=0
    """
    diskimage_path = os.path.join(disk_images_directory, _mount_name(path))
    disk_size = _overlay_disk_size_old()

    api.current_logger().debug('Attempting to create disk image with size %d MiB at %s', disk_size, diskimage_path)
    utils.call_with_failure_hint(
        cmd=['/bin/dd', 'if=/dev/zero', 'of={}'.format(diskimage_path), 'bs=1M', 'count={}'.format(disk_size)],
        hint='Please ensure that there is enough diskspace in {} at least {} MiB are needed'.format(
            diskimage_path, disk_size)
    )

    api.current_logger().debug('Creating ext4 filesystem in disk image at %s', diskimage_path)
    try:
        utils.call_with_oserror_handled(cmd=['/sbin/mkfs.ext4', '-F', diskimage_path])
    except CalledProcessError as e:
        api.current_logger().error('Failed to create ext4 filesystem in %s', exc_info=True)
        raise StopActorExecutionError(
            message=str(e)
        )

    return diskimage_path


def _prepare_required_mounts_old(scratch_dir, mounts_dir, mount_points, xfs_info):
    result = {
        mount_point.fs_file: mounting.NullMount(
            _mount_dir(mounts_dir, mount_point.fs_file)) for mount_point in mount_points
    }

    if not xfs_info.mountpoints_without_ftype:
        return result

    space_needed = _overlay_disk_size_old() * len(xfs_info.mountpoints_without_ftype)
    disk_images_directory = os.path.join(scratch_dir, 'diskimages')

    # Ensure we cleanup old disk images before we check for space constraints.
    run(['rm', '-rf', disk_images_directory])
    _create_diskimages_dir_old(scratch_dir, disk_images_directory)
    _ensure_enough_diskimage_space_old(space_needed, scratch_dir)

    mount_names = [mount_point.fs_file for mount_point in mount_points]

    # TODO(pstodulk): this (adding rootfs into the set always) is hotfix for
    # bz #1911802 (not ideal one..). The problem occurs one rootfs is ext4 fs,
    # but /var/lib/leapp/... is under XFS without ftype; In such a case we can
    # see still the very same problems as before. But letting you know that
    # probably this is not the final solution, as we could possibly see the
    # same problems on another partitions too (needs to be tested...). However,
    # it could fit for now until we provide the complete solution around XFS
    # workarounds (including management of required spaces for virtual FSs per
    # mountpoints - without that, we cannot fix this properly)
    for mountpoint in set(xfs_info.mountpoints_without_ftype + ['/']):
        if mountpoint in mount_names:
            image = _create_mount_disk_image_old(disk_images_directory, mountpoint)
            result[mountpoint] = mounting.LoopMount(source=image, target=_mount_dir(mounts_dir, mountpoint))
    return result



import errno
import os

from leapp.libraries.common import multipathutil
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import (
    CopyFile,
    DistributionSignedRPM,
    MultipathConfFacts,
    MultipathConfig,
    MultipathConfigOption,
    TargetUserSpaceUpgradeTasks
)


def _change_existing_option(curr_options, opt_name, opt_value):
    for option in curr_options:
        if option.name == opt_name:
            option.value = opt_value  # latest value is used
            return True
    return False


def _add_options(curr_options, new_options):
    ignore = ['hardware_handler', 'pg_timeout', 'product', 'unpriv_sgio',
              'product_blacklist', 'revision', 'vendor']
    for opt_name, opt_value in new_options:
        if opt_name in ignore:
            continue
        if opt_name == 'detect_path_checker':
            opt_name = 'detect_checker'
        if not _change_existing_option(curr_options, opt_name, opt_value):
            curr_options.append(MultipathConfigOption(name=opt_name,
                                                      value=opt_value))


def _remove_qinp(value):
    items = value.split()
    if items == [] or not items[0].isdigit():
        return value
    nr_features = int(items[0])
    if nr_features != len(items) - 1:
        return value
    try:
        items.remove('queue_if_no_path')
    except ValueError:
        return value
    items[0] = str(nr_features - 1)
    return ' '.join(items)


def _fix_qinp_options(options):
    have_npr = False
    need_npr = False
    for option in options:
        if option.name == 'features' and 'queue_if_no_path' in option.value:
            option.value = _remove_qinp(option.value)
            need_npr = True
        if option.name == 'no_path_retry':
            have_npr = True
    if need_npr and not have_npr:
        options.append(MultipathConfigOption(name='no_path_retry',
                                             value='queue'))


def _options_match(overrides, all_devs):
    if overrides == 'detect_path_checker' and all_devs == 'detect_checker':
        return True
    if overrides in ('path_checker', 'checker') and \
            all_devs in ('path_checker', 'checker'):
        return True
    if overrides == all_devs:
        return True
    return False


def _filter_options(all_dev_options, overrides_options):
    for name, value in overrides_options:
        if name == 'features' and 'queue_if_no_path' in value:
            overrides_options.append(('no_path_retry', 'queue'))
            break
    for name, _value in overrides_options:
        for option in all_dev_options:
            if _options_match(name, option.name):
                all_dev_options.remove(option)
                break


def _parse_config(path):
    contents = multipathutil.read_config(path)
    if contents is None:
        return None
    conf = MultipathConfig(pathname=path)
    conf.all_devs_options = []
    section = None
    in_subsection = False
    device_options = []
    overrides_options = []
    in_all_devs = False
    for line in contents.split('\n'):
        try:
            data = multipathutil.LineData(line, section, in_subsection)
        except ValueError:
            continue
        if data.type == data.TYPE_BLANK:
            continue
        if data.type == data.TYPE_SECTION_END:
            if in_subsection:
                in_subsection = False
                if in_all_devs:
                    _add_options(conf.all_devs_options, device_options)
                in_all_devs = False
                device_options = []
            elif section:
                section = None
            continue
        if data.type == data.TYPE_SECTION_START:
            if not section:
                section = data.section
            elif not in_subsection:
                in_subsection = True
            continue
        if data.type != data.TYPE_OPTION:
            continue
        if section == 'defaults':
            if data.option in ('path_checker', 'checker'):
                conf.default_path_checker = data.value
            elif data.option == 'config_dir':
                conf.config_dir = data.value
            elif data.option == 'retain_attached_hw_handler':
                conf.default_retain_hwhandler = data.is_enabled()
            elif data.option == 'detect_prio':
                conf.default_detect_prio = data.is_enabled()
            elif data.option == 'detect_path_checker':
                conf.default_detect_checker = data.is_enabled()
            elif data.option == 'reassign_maps':
                conf.reassign_maps = data.is_enabled()
            elif data.option == 'hw_str_match':
                conf.hw_str_match_exists = True
            elif data.option == 'ignore_new_boot_devs':
                conf.ignore_new_boot_devs_exists = True
            elif data.option == 'new_bindings_in_boot':
                conf.new_bindings_in_boot_exists = True
        if section == 'devices' and in_subsection:
            if data.option == 'all_devs' and data.is_enabled():
                conf.all_devs_section_exists = True
                in_all_devs = True
            else:
                device_options.append((data.option, data.value))
        if section == 'overrides':
            if data.option == 'hardware_handler':
                conf.overrides_hwhandler_exists = True
            elif data.option == 'pg_timeout':
                conf.overrides_pg_timeout_exists = True
            else:
                overrides_options.append((data.option, data.value))
        if data.option == 'unpriv_sgio':
            conf.unpriv_sgio_exists = True
        if data.option == 'detect_path_checker':
            conf.detect_path_checker_exists = True
        if data.option == 'features' and 'queue_if_no_path' in data.value:
            conf.queue_if_no_path_exists = True

    if in_subsection and in_all_devs:
        _add_options(conf.all_devs_options, device_options)
    _fix_qinp_options(conf.all_devs_options)
    _filter_options(conf.all_devs_options, overrides_options)
    return conf


def _parse_config_dir(config_dir):
    res = []
    try:
        for config_file in sorted(os.listdir(config_dir)):
            path = os.path.join(config_dir, config_file)
            if not path.endswith('.conf'):
                continue
            conf = _parse_config(path)
            if conf:
                res.append(conf)
    except OSError as e:
        if e.errno == errno.ENOENT:
            api.current_logger().debug('Multipath conf directory ' +
                                       '"{}" doesn\'t exist'.format(config_dir))
        else:
            api.current_logger().warning('Failed to read multipath config ' +
                                         'directory ' +
                                         '"{}": {}'.format(config_dir, e))
    return res


def is_processable():
    res = has_package(DistributionSignedRPM, 'device-mapper-multipath')
    if not res:
        api.current_logger().debug('device-mapper-multipath is not installed.')
    return res


def get_multipath_conf_facts(config_file='/etc/multipath.conf'):
    res_configs = []
    conf = _parse_config(config_file)
    if not conf:
        return None
    res_configs.append(conf)
    if conf.config_dir:
        res_configs.extend(_parse_config_dir(conf.config_dir))
    else:
        res_configs.extend(_parse_config_dir('/etc/multipath/conf.d'))
    return MultipathConfFacts(configs=res_configs)


def produce_copy_to_target_task():
    """
    Produce task to copy files into the target userspace

    The multipath configuration files are needed when the upgrade init ramdisk
    is generated to ensure we are able to boot into the upgrade environment
    and start the upgrade process itself. By this msg it's told that these
    files/dirs will be available when the upgrade init ramdisk is generated.

    See TargetUserSpaceUpgradeTasks and UpgradeInitramfsTasks for more info.
    """
    # TODO(pstodulk): move the function to the multipathconfcheck actor
    # and get rid of the hardcoded stuff.
    # - The current behaviour looks from the user POV same as before this
    # * commit. I am going to keep the proper fix for additional PR as we do
    # * not want to make the current PR even more complex than now and the solution
    # * is not so trivial.
    # - As well, I am missing some information around xDR devices, which are
    # * possibly not handled correctly (maybe missing some executables?..)
    # * Update: practically we do not have enough info about xDR drivers, but
    # * discussed with Ben Marzinski, as the multipath dracut module includes
    # * the xDR utils stuff, we should handle it in the same way.
    # * See xdrgetuid, xdrgetinfo (these two utils are now missing in our initramfs)
    copy_files = []
    for fname in ['/etc/multipath.conf', '/etc/multipath', '/etc/xdrdevices.conf']:
        if os.path.exists(fname):
            copy_files.append(CopyFile(src=fname))

    if copy_files:
        api.produce(TargetUserSpaceUpgradeTasks(copy_files=copy_files))



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class VsftpdConfig(Model):
    """
    Model representing some aspects of a vsftpd configuration file.

    The attributes representing the state of configuration options are nullable, so that
    they can represent the real state of the option in the file: if an option is set to "YES"
    in the configuration file, the corresponding attribute is set to True; if the option
    is set to NO, the attribute is set to False; if the option is not present in the config
    file at all, the attribute is set to None.
    """
    topic = SystemInfoTopic

    path = fields.String()
    """Path to the vsftpd configuration file"""
    strict_ssl_read_eof = fields.Nullable(fields.Boolean())
    """Represents the state of the strict_ssl_read_eof option in the config file"""
    tcp_wrappers = fields.Nullable(fields.Boolean())
    """Represents the state of the tcp_wrappers option in the config file"""


class VsftpdFacts(Model):
    topic = SystemInfoTopic

    default_config_hash = fields.Nullable(fields.String())
    """SHA1 hash of the /etc/vsftpd/vsftpd.conf file, if it exists, None otherwise"""
    configs = fields.List(fields.Model(VsftpdConfig))
    """List of vsftpd configuration files"""



import os

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import systemd
from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import SystemdBrokenSymlinksSource, SystemdBrokenSymlinksTarget, SystemdServicesInfoSource

_INSTALLATION_CHANGED_EL8 = ['rngd.service', 'sysstat.service']
_INSTALLATION_CHANGED_EL9 = []


def _get_installation_changed_units():
    version = get_target_major_version()
    if version == '8':
        return _INSTALLATION_CHANGED_EL8
    if version == '9':
        return _INSTALLATION_CHANGED_EL9

    return []


def _service_enabled_source(service_info, name):
    service_file = next((s for s in service_info.service_files if s.name == name), None)
    return service_file and service_file.state == 'enabled'


def _is_unit_enabled(unit):
    try:
        ret = run(['systemctl', 'is-enabled', unit], split=True)['stdout']
        return ret and ret[0] == 'enabled'
    except (OSError, CalledProcessError):
        return False


def _handle_newly_broken_symlinks(symlinks, service_info):
    for symlink in symlinks:
        unit = os.path.basename(symlink)
        try:
            if not _is_unit_enabled(unit):
                # removes the broken symlink
                systemd.disable_unit(unit)
            elif _service_enabled_source(service_info, unit) and _is_unit_enabled(unit):
                # removes the old symlinks and creates the new ones
                systemd.reenable_unit(unit)
        except CalledProcessError:
            # TODO(mmatuska): Produce post-upgrade report: failed to handle broken symlink (and suggest a fix?)
            pass


def _handle_bad_symlinks(service_files):
    install_changed_units = _get_installation_changed_units()
    potentially_bad = [s for s in service_files if s.name in install_changed_units]

    for unit_file in potentially_bad:
        if unit_file.state == 'enabled' and _is_unit_enabled(unit_file.name):
            systemd.reenable_unit(unit_file.name)


def process():
    service_info_source = next(api.consume(SystemdServicesInfoSource), None)
    if not service_info_source:
        raise StopActorExecutionError("Expected SystemdServicesInfoSource message, but got None")

    source_info = next(api.consume(SystemdBrokenSymlinksSource), None)
    target_info = next(api.consume(SystemdBrokenSymlinksTarget), None)

    if source_info and target_info:
        newly_broken = []
        newly_broken = [s for s in target_info.broken_symlinks if s not in source_info.broken_symlinks]
        if not newly_broken:
            return

        _handle_newly_broken_symlinks(newly_broken, service_info_source)

    _handle_bad_symlinks(service_info_source.service_files)



import os

import pytest

from leapp.libraries.actor import scanzfcp
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked, produce_mocked
from leapp.models import CopyFile, TargetUserSpaceUpgradeTasks, UpgradeInitramfsTasks


def test_zfcp_exists(monkeypatch):
    monkeypatch.setattr(scanzfcp.api, 'current_actor', CurrentActorMocked(arch=architecture.ARCH_S390X))
    monkeypatch.setattr(scanzfcp.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(scanzfcp.api, 'produce', produce_mocked())
    monkeypatch.setattr(os.path, 'isfile', lambda dummy: True)
    scanzfcp.process()
    assert not scanzfcp.api.current_logger.infomsg
    assert scanzfcp.api.produce.called == 2
    tusut_flag = False
    uit_flag = False
    for msg in scanzfcp.api.produce.model_instances:
        if isinstance(msg, TargetUserSpaceUpgradeTasks):
            assert [CopyFile(src=scanzfcp.ZFCP_CONF)] == msg.copy_files
            assert msg.install_rpms == ['s390utils-core']
            tusut_flag = True
        elif isinstance(msg, UpgradeInitramfsTasks):
            assert [scanzfcp.ZFCP_CONF] == msg.include_files
            uit_flag = True
    assert tusut_flag and uit_flag


def test_zfcp_not_found(monkeypatch):
    monkeypatch.setattr(scanzfcp.api, 'current_actor', CurrentActorMocked(arch=architecture.ARCH_S390X))
    monkeypatch.setattr(scanzfcp.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(scanzfcp.os.path, 'isfile', lambda dummy: False)
    monkeypatch.setattr(scanzfcp.api, 'produce', produce_mocked())
    scanzfcp.process()
    assert scanzfcp.api.current_logger.infomsg
    assert scanzfcp.api.produce.called == 1
    assert len(scanzfcp.api.produce.model_instances) == 1
    assert isinstance(scanzfcp.api.produce.model_instances[0], TargetUserSpaceUpgradeTasks)
    assert scanzfcp.api.produce.model_instances[0].install_rpms == ['s390utils-core']
    assert not scanzfcp.api.produce.model_instances[0].copy_files


@pytest.mark.parametrize('isfile', [True, False])
@pytest.mark.parametrize('arch', [
    architecture.ARCH_X86_64,
    architecture.ARCH_ARM64,
    architecture.ARCH_PPC64LE,
])
def test_non_ibmz_arch(monkeypatch, isfile, arch):
    monkeypatch.setattr(scanzfcp.api, 'current_actor', CurrentActorMocked(arch=arch))
    monkeypatch.setattr(scanzfcp.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(scanzfcp.api, 'produce', produce_mocked())
    monkeypatch.setattr(os.path, 'isfile', lambda dummy: isfile)
    scanzfcp.process()
    assert not scanzfcp.api.current_logger.infomsg
    assert not scanzfcp.api.produce.called



from leapp.actors import Actor
from leapp.libraries.actor import systemfacts
from leapp.libraries.common.config import architecture
from leapp.models import (
    ActiveKernelModulesFacts,
    DefaultGrubInfo,
    FirewallsFacts,
    FirmwareFacts,
    GroupsFacts,
    GrubCfgBios,
    Report,
    RepositoriesFacts,
    SELinuxFacts,
    SysctlVariablesFacts,
    UsersFacts
)
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class SystemFactsActor(Actor):
    """
    Provides data about many facts from system.

    After collecting data from multiple tools, messages with relevant data will be produced to
    describe:
      - Sysctl variables;
      - Active Linux Kernel Modules;
      - Users;
      - Groups;
      - Package repositories;
      - SELinux status;
      - Firewalls status.
    """

    name = 'system_facts'
    consumes = ()
    produces = (
        SysctlVariablesFacts,
        ActiveKernelModulesFacts,
        UsersFacts,
        GroupsFacts,
        RepositoriesFacts,
        SELinuxFacts,
        FirewallsFacts,
        FirmwareFacts,
        DefaultGrubInfo,
        GrubCfgBios,
        Report
    )
    tags = (IPUWorkflowTag, FactsPhaseTag,)

    def process(self):
        self.produce(systemfacts.get_sysctls_status())
        self.produce(systemfacts.get_active_kernel_modules_status(self.log))
        self.produce(systemfacts.get_system_users_status())
        self.produce(systemfacts.get_system_groups_status())
        self.produce(systemfacts.get_repositories_status())
        self.produce(systemfacts.get_selinux_status())
        self.produce(systemfacts.get_firewalls_status())
        self.produce(systemfacts.get_firmware())

        if not architecture.matches_architecture(architecture.ARCH_S390X):
            self.produce(systemfacts.get_default_grub_conf())

        bios_grubcfg_details = systemfacts.get_bios_grubcfg_details()
        if bios_grubcfg_details:
            self.produce(bios_grubcfg_details)



from leapp.actors import Actor
from leapp.libraries.actor.checkmemcached import check_memcached
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckMemcached(Actor):
    """
    Check for incompatible changes in memcached configuration.

    Warn that memcached in RHEL8 no longer listens on the UDP port by default
    and the default service configuration binds memcached to the loopback
    interface.
    """

    name = 'check_memcached'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        check_memcached(has_package(DistributionSignedRPM, 'memcached'))



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class BrlttyMigrationDecision(Model):
    topic = SystemInfoTopic
    migrate_file = fields.String()
    migrate_bt = fields.Boolean()
    migrate_espeak = fields.Boolean()



from leapp.actors import Actor
from leapp.libraries.actor.workaround import apply_python3_workaround
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class PreparePythonWorkround(Actor):
    """
    Prepare environment to be able to run leapp with Python3 in initrd.

    These are the current necessary steps to be able to run Leapp with Python3.
    Basically, we create directory (now /root/tmp_leapp_py3/). We will put
    symlinks inside which will point to leapp python packages. Additionally,
    we create new script that will import expected modules and run leapp again.
    """

    name = 'prepare_python_workround'
    consumes = ()
    produces = ()
    tags = (IPUWorkflowTag, RPMUpgradePhaseTag)

    def process(self):
        apply_python3_workaround()



from leapp.actors import Actor
from leapp.libraries.actor import detectkerneldrivers
from leapp.models import ActiveKernelModulesFacts, DetectedDeviceOrDriver, DeviceDriverDeprecationData
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class DetectKernelDrivers(Actor):
    """
    Matches all currently loaded kernel drivers against known deprecated and removed drivers.
    """

    name = 'detect_kernel_drivers'
    consumes = (DeviceDriverDeprecationData, ActiveKernelModulesFacts)
    produces = (DetectedDeviceOrDriver,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        detectkerneldrivers.process()



#!/usr/bin/env python
#
# Simplified parsing of bind configuration, with include support and nested sections.

from __future__ import print_function

import re
import string


class ConfigParseError(Exception):
    """Generic error when parsing config file."""

    def __init__(self, error=None, parent=None):
        # IOError on python3 includes path, on python2 it does not
        message = "Cannot open the configuration file \"{path}\": {error}".format(
                    path=error.filename, error=str(error))
        if parent:
            message += "; included from \"{0}\"".format(parent)
        super(ConfigParseError, self).__init__(message)
        self.error = error
        self.parent = parent
    pass


class ConfigFile(object):
    """Representation of single configuration file and its contents."""
    def __init__(self, path):
        """Load config file contents from path.

        :param path: Path to file
        """
        self.path = path
        self.load(path)
        self.status = None

    def __str__(self):
        return self.buffer

    def __repr__(self):
        return 'ConfigFile {0} ({1})'.format(
                self.path, self.buffer)

    def load(self, path):
        with open(path, 'r') as f:
            self.buffer = self.original = f.read()

    def is_modified(self):
        return self.original == self.buffer

    def root_section(self):
        return ConfigSection(self, None, 0, len(self.buffer))


class MockConfig(ConfigFile):
    """Configuration file with contents defined on constructor.

       Intended for testing the library.
    """
    DEFAULT_PATH = '/etc/named/mock.conf'

    def __init__(self, contents, path=DEFAULT_PATH):
        self.original = contents
        super(MockConfig, self).__init__(path)

    def load(self, path):
        self.buffer = self.original


class ConfigSection(object):
    """Representation of section or key inside single configuration file.

    Section means statement, block, quoted string or any similar."""

    TYPE_BARE = 1
    TYPE_QSTRING = 2
    TYPE_BLOCK = 3
    TYPE_IGNORED = 4  # comments and whitespaces

    def __init__(self, config, name=None, start=None, end=None, kind=None, parser=None):
        """
        :param config: config file inside which is this section
        :type config: ConfigFile
        :param kind: type of this section
        """
        self.config = config
        self.name = name
        self.start = start
        self.end = end
        self.ctext = self.original_value()   # a copy for modification
        self.parser = parser
        if kind is None:
            if self.config.buffer.startswith('{', self.start):
                self.kind = self.TYPE_BLOCK
            elif self.config.buffer.startswith('"', self.start):
                self.kind = self.TYPE_QSTRING
            else:
                self.kind = self.TYPE_BARE
        else:
            self.kind = kind
        self.statements = []

    def __repr__(self):
        text = self.value()
        path = self.config.path
        return 'ConfigSection#{kind}({path}:{start}-{end}: "{text}")'.format(
            path=path, start=self.start, end=self.end,
            text=text, kind=self.kind
        )

    def __str__(self):
        return self.value()

    def copy(self):
        return ConfigSection(self.config, self.name, self.start, self.end, self.kind)

    def type(self):
        return self.kind

    def value(self):
        return self.ctext

    def original_value(self):
        return self.config.buffer[self.start:self.end+1]

    def invalue(self):
        """Return just inside value of blocks and quoted strings."""
        t = self.type()
        if t in (self.TYPE_QSTRING, self.TYPE_BLOCK):
            return self.ctext[1:-1]
        return self.value()

    def children(self, comments=False):
        """Return list of items inside this section."""
        start = self.start
        if self.type() == self.TYPE_BLOCK:
            start += 1
        return list(IscIterator(self.parser, self, comments, start))

    def serialize(self):
        return self.value()


class IscIterator(object):
    """Iterator for walking over parsed configuration.

       Creates sequence of ConfigSection objects for a given file.
       That means a stream of objects.
    """

    def __init__(self, parser, section, comments=False, start=None):
        """Create iterator.

        :param comments: Include comments and whitespaces
        :param start: Index for starting, None means beginning of section
        """
        self.parser = parser
        self.section = section
        self.current = None
        self.key_wanted = True
        self.comments = comments
        self.waiting = None
        if start is None:
            start = section.start
        self.start = start

    def __iter__(self):
        self.current = None
        self.key_wanted = True
        self.waiting = None
        return self

    def __next__(self):
        index = self.start
        cfg = self.section.config
        if self.waiting:
            self.current = self.waiting
            self.waiting = None
            return self.current
        if self.current is not None:
            index = self.current.end+1
        if self.key_wanted:
            val = self.parser.find_next_key(cfg, index, self.section.end)
            self.key_wanted = False
        else:
            val = self.parser.find_next_val(cfg, None, index, self.section.end, end_report=True)
            if val is not None and val.value() in self.parser.CHAR_DELIM:
                self.key_wanted = True
        if val is None:
            if self.current is not None and self.current.end < self.section.end and self.comments:
                self.current = ConfigSection(self.section.config, None,
                                             index, self.section.end, ConfigSection.TYPE_IGNORED)
                return self.current
            raise StopIteration
        if index != val.start and self.comments:
            # Include comments and spaces as ignored section
            self.waiting = val
            val = ConfigSection(val.config, None, index, val.start-1, ConfigSection.TYPE_IGNORED)

        self.current = val
        return val

    next = __next__  # Python2 compat


class IscVarIterator(object):
    """Iterator for walking over parsed configuration.

       Creates sequence of ConfigVariableSection objects for a given
       file or section.
    """

    def __init__(self, parser, section, comments=False, start=None):
        """Create iterator."""
        self.parser = parser
        self.section = section
        self.iter = IscIterator(parser, section, comments, start)

    def __iter__(self):
        return self

    def __next__(self):
        vl = []
        try:
            statement = next(self.iter)
            while statement:
                vl.append(statement)
                if self.parser.is_terminal(statement):
                    return ConfigVariableSection(vl, None, parent=self.section)
                statement = next(self.iter)
        except StopIteration:
            if vl:
                return ConfigVariableSection(vl, None, parent=self.section)
        raise StopIteration

    next = __next__  # Python2 compat


class ConfigVariableSection(ConfigSection):
    """Representation for key and values of variable length.

    Intended for view and zone.
    """

    def __init__(self, sectionlist, name, zone_class=None, parent=None, parser=None):
        """Creates variable block for zone or view.

        :param sectionlist: list of ConfigSection, obtained from IscConfigParser.find_values()
        """
        last = next(reversed(sectionlist))
        first = sectionlist[0]
        self.values = sectionlist
        super(ConfigVariableSection, self).__init__(
            first.config, name, start=first.start, end=last.end, parser=parser
        )
        if name is None:
            try:
                self.name = self.var(1).invalue()
            except IndexError:
                pass
        # For optional dns class, like IN or CH
        self.zone_class = zone_class
        self.parent = parent

    def key(self):
        if self.zone_class is None:
            return self.name
        return self.zone_class + '_' + self.name

    def firstblock(self):
        """Return first block section in this tool."""
        return self.vartype(0, self.TYPE_BLOCK)

    def var(self, i):
        """Return value by index, ignore spaces."""
        n = 0
        for v in self.values:
            if v.type() != ConfigSection.TYPE_IGNORED:
                if n == i:
                    return v
                n += 1
        raise IndexError

    def vartype(self, i, vtype):
        n = 0
        for v in self.values:
            if v.type() == vtype:
                if n == i:
                    return v
                n += 1
        raise IndexError

    def serialize(self):
        s = ''
        for v in self.values:
            s += v.serialize()
        return s

    def serialize_skip(self, replace_ignored=None):
        """
        Create single string from section, but skip whitespace on start.

        :type section: ConfigVariableSection
        :param replace_ignored: Specify replaced text for whitespace

        Allows normalizing with replace ignored sections.
        Is intended to strip possible comments between parts.
        """
        s = ''
        nonwhite = None
        for v in self.values:
            if nonwhite is None:
                if v.type() != self.TYPE_IGNORED:
                    nonwhite = v
                    s += v.serialize()
            elif replace_ignored is not None and v.type() == self.TYPE_IGNORED:
                s += replace_ignored
            else:
                s += v.serialize()
        return s


class ModifyState(object):
    """Object keeping state of modifications when walking configuration file statements.

    It would keep modified configuration file and position of last found statement.
    """

    def __init__(self):
        self.value = ''
        self.lastpos = 0

    def append_before(self, section):
        """Appends content from last seen section to beginning of current one.

        It adds also whitespace on beginning of statement,
        which is usually not interesting for any changes.

        :type section: ConfigVariableSection
        """

        end = section.start
        first = section.values[0]
        if first.type() == first.TYPE_IGNORED:
            end = first.end
        cfg = section.config.buffer
        self.value += cfg[self.lastpos:end+1]
        self.lastpos = end+1

    def move_after(self, section):
        """Set position to the end of section."""
        self.lastpos = section.end+1

    def finish(self, section):
        """Append remaining part of file to modified state."""
        if self.lastpos < section.end:
            self.value += section.config.buffer[self.lastpos:section.end+1]
            self.lastpos = section.end

    def content(self):
        """Get content of (modified) section.

        Would be valid after finish() was called.
        """
        return self.value

    @staticmethod
    def callback_comment_out(section, state):
        """parser.walk callback for commenting out the section."""
        state.append_before(section)
        state.value += '/* ' + section.serialize_skip(' ') + ' */'
        state.move_after(section)

    @staticmethod
    def callback_remove(section, state):
        """parser.walk callback for skipping a section."""
        state.append_before(section)
        state.move_after(section)


# Main parser class
class IscConfigParser(object):
    """Parser file with support of included files.

    Reads ISC BIND configuration file and tries to skip commented blocks, nested sections and similar stuff.
    Imitates what isccfg does in native code, but without any use of native code.
    """

    CONFIG_FILE = "/etc/named.conf"
    FILES_TO_CHECK = []

    CHAR_DELIM = ";"  # Must be single character
    CHAR_CLOSING = CHAR_DELIM + "})]"
    CHAR_CLOSING_WHITESPACE = CHAR_CLOSING + string.whitespace
    CHAR_KEYWORD = string.ascii_letters + string.digits + '-_.:'
    CHAR_STR_OPEN = '"'

    def __init__(self, config=None):
        """Construct parser.

        :param config: path to file or already loaded ConfigFile instance

        Initialize contents from path to real config or already loaded ConfigFile class.
        """
        if isinstance(config, ConfigFile):
            self.FILES_TO_CHECK = [config]
            self.load_included_files()
        elif config is not None:
            self.load_config(config)

    #
    # function for parsing of config files
    #
    def is_comment_start(self, istr, index=0):
        if istr[index] == "#" or (
                index+1 < len(istr) and istr[index:index+2] in ["//", "/*"]):
            return True
        return False

    def _find_end_of_comment(self, istr, index=0):
        """Returns index where the comment ends.

        :param istr: input string
        :param index: begin search from the index; from the start by default

        Support usual comments till the end of line (//, #) and block comment
        like (/* comment */). In case that index is outside of the string or end
        of the comment is not found, return -1.

        In case of block comment, returned index is position of slash after star.
        """
        length = len(istr)

        if index >= length or index < 0:
            return -1

        if istr[index] == "#" or istr[index:].startswith("//"):
            return istr.find("\n", index)

        if index+2 < length and istr[index:index+2] == "/*":
            res = istr.find("*/", index+2)
            if res != -1:
                return res + 1

        return -1

    def is_opening_char(self, c):
        return c in "\"'{(["

    def _remove_comments(self, istr, space_replace=False):
        """Removes all comments from the given string.

        :param istr: input string
        :param space_replace When true, replace comments with spaces. Skip them by default.
        :return: istr without comments
        """

        ostr = ""

        length = len(istr)
        index = 0

        while index < length:
            if self.is_comment_start(istr, index):
                index = self._find_end_of_comment(istr, index)
                if index == -1:
                    index = length
                if space_replace:
                    ostr = ostr.ljust(index)
                if index < length and istr[index] == "\n":
                    ostr += "\n"
            elif istr[index] in self.CHAR_STR_OPEN:
                end_str = self._find_closing_char(istr, index)
                if end_str == -1:
                    ostr += istr[index:]
                    break
                ostr += istr[index:end_str+1]
                index = end_str
            else:
                ostr += istr[index]
            index += 1

        return ostr

    def _replace_comments(self, istr):
        """Replaces all comments by spaces in the given string.

        :param istr: input string
        :returns: string of the same length with comments replaced
        """
        return self._remove_comments(istr, True)

    def find_next_token(self, istr, index=0, end_index=-1, end_report=False):
        """
        Return index of another interesting token or -1 when there is not next.

        :param istr: input string
        :param index: begin search from the index; from the start by default
        :param end_index: stop searching at the end_index or end of the string

        In case that initial index contains already some token, skip to another.
        But when searching starts on whitespace or beginning of the comment,
        choose the first one.

        The function would be confusing in case of brackets, but content between
        brackets is not evaluated as new tokens.
        E.g.:

        "find { me };"      : 5
        " me"               : 1
        "find /* me */ me " : 13
        "/* me */ me"       : 9
        "me;"               : 2
        "{ me }; me"        : 6
        "{ me }  me"        : 8
        "me }  me"          : 3
        "}} me"             : 1
        "me"                : -1
        "{ me } "           : -1
        """
        length = len(istr)
        if length < end_index or end_index < 0:
            end_index = length

        if index >= end_index or index < 0:
            return -1

        # skip to the end of the current token
        if istr[index] == '\\':
            index += 2
        elif self.is_opening_char(istr[index]):
            index = self._find_closing_char(istr, index, end_index)
            if index != -1:
                index += 1
        elif self.is_comment_start(istr, index):
            index = self._find_end_of_comment(istr, index)
            if index != -1:
                index += 1
        elif istr[index] not in self.CHAR_CLOSING_WHITESPACE:
            # so we have to skip to the end of the current token
            index += 1
            while index < end_index:
                if (istr[index] in self.CHAR_CLOSING_WHITESPACE
                        or self.is_comment_start(istr, index)
                        or self.is_opening_char(istr[index])):
                    break
                index += 1
        elif end_report and istr[index] in self.CHAR_DELIM:
            # Found end of statement. Report delimiter
            return index
        elif istr[index] in self.CHAR_CLOSING:
            index += 1

        # find next token (can be already under the current index)
        while 0 <= index < end_index:
            if istr[index] == '\\':
                index += 2
                continue
            if self.is_comment_start(istr, index):
                index = self._find_end_of_comment(istr, index)
                if index == -1:
                    break
            elif self.is_opening_char(istr[index]) or istr[index] not in string.whitespace:
                return index
            index += 1
        return -1

    def _find_closing_char(self, istr, index=0, end_index=-1):
        """
        Returns index of equivalent closing character.

        :param istr: input string

        It's similar to the "find" method that returns index of the first character
        of the searched character or -1. But in this function the corresponding
        closing character is looked up, ignoring characters inside strings
        and comments. E.g. for
            "(hello (world) /* ) */ ), he would say"
        index of the third ")" is returned.
        """
        important_chars = {  # TODO: should be that rather global var?
            "{": "}",
            "(": ")",
            "[": "]",
            "\"": "\"",
            self.CHAR_DELIM: None,
            }
        length = len(istr)
        if 0 <= end_index < length:
            length = end_index

        if length < 2:
            return -1

        if index >= length or index < 0:
            return -1

        closing_char = important_chars.get(istr[index], self.CHAR_DELIM)
        if closing_char is None:
            return -1

        isString = istr[index] in "\""
        index += 1
        curr_c = ""
        while index < length:
            curr_c = istr[index]
            if curr_c == '//':
                index += 2
            elif self.is_comment_start(istr, index) and not isString:
                index = self._find_end_of_comment(istr, index)
                if index == -1:
                    return -1
            elif not isString and self.is_opening_char(curr_c):
                deep_close = self._find_closing_char(istr[index:])
                if deep_close == -1:
                    break
                index += deep_close
            elif curr_c == closing_char:
                if curr_c == self.CHAR_DELIM:
                    index -= 1
                return index
            index += 1

        return -1

    def find_key(self, istr, key, index=0, end_index=-1, only_first=True):
        """
        Return index of the key or -1.

        :param istr: input string; it could be whole file or content of a section
        :param key: name of the searched key in the current scope
        :param index: start searching from the index
        :param end_index: stop searching at the end_index or end of the string

        Function is not recursive. Searched key has to be in the current scope.
        Attention:

        In case that input string contains data outside of section by mistake,
        the closing character is ignored and the key outside of scope could be
        found. Example of such wrong input could be:
              key1 "val"
              key2 { key-ignored "val-ignored" };
            };
            controls { ... };
        In this case, the key "controls" is outside of original scope. But for this
        cases you can set end_index to value, where searching should end. In case
        you set end_index higher then length of the string, end_index will be
        automatically corrected to the end of the input string.
        """
        length = len(istr)
        keylen = len(key)

        if length < end_index or end_index < 0:
            end_index = length

        if index >= end_index or index < 0:
            return -1

        while index != -1:
            if istr.startswith(key, index):
                if index+keylen < end_index and istr[index+keylen] not in self.CHAR_KEYWORD:
                    # key has been found
                    return index

            while not only_first and index != -1 and istr[index] != self.CHAR_DELIM:
                index = self.find_next_token(istr, index)
            index = self.find_next_token(istr, index)

        return -1

    def find_next_key(self, cfg, index=0, end_index=-1, end_report=False):
        """Modernized variant of find_key.

        :type cfg: ConfigFile
        :param index: Where to start search
        :rtype: ConfigSection

        Searches for first place of bare keyword, without quotes or block.
        """
        istr = cfg.buffer
        length = len(istr)

        if length < end_index or end_index < 0:
            end_index = length

        if index > end_index or index < 0:
            raise IndexError("Invalid cfg index")

        while index != -1:
            keystart = index
            while index < end_index and istr[index] in self.CHAR_KEYWORD:
                index += 1

            if index >= end_index:
                break

            if keystart < index <= end_index and istr[index] not in self.CHAR_KEYWORD:
                # key has been found
                return ConfigSection(cfg, istr[keystart:index], keystart, index-1)
            if istr[index] in self.CHAR_DELIM:
                return ConfigSection(cfg, istr[index], index, index)

            index = self.find_next_token(istr, index, end_index, end_report)

        return None

    def find_next_val(self, cfg, key=None, index=0, end_index=-1, end_report=False):
        """Find following token.

        :param cfg: input token
        :type cfg: ConfigFile
        :returns: ConfigSection object or None
        :rtype: ConfigSection
        """
        start = self.find_next_token(cfg.buffer, index, end_index, end_report)
        if start < 0:
            return None
        if end_index < 0:
            end_index = len(cfg.buffer)
        # remains = cfg.buffer[start:end_index]
        if not self.is_opening_char(cfg.buffer[start]):
            return self.find_next_key(cfg, start, end_index, end_report)

        end = self._find_closing_char(cfg.buffer, start, end_index)
        if end == -1 or (0 < end_index < end):
            return None
        return ConfigSection(cfg, key, start, end)

    def find_val(self, cfg, key, index=0, end_index=-1):
        """Find value of keyword specified by key.

        :param cfg: ConfigFile
        :param key: name of searched key (str)
        :param index: start of search in cfg (int)
        :param end_index: end of search in cfg (int)
        :returns: ConfigSection object or None
        :rtype: ConfigSection
        """
        if not isinstance(cfg, ConfigFile):
            raise TypeError("cfg must be ConfigFile parameter")

        if end_index < 0:
            end_index = len(cfg.buffer)
        key_start = self.find_key(cfg.buffer, key, index, end_index)
        if key_start < 0 or key_start+len(key) >= end_index:
            return None
        return self.find_next_val(cfg, key, key_start+len(key), end_index)

    def find_val_section(self, section, key):
        """Find value of keyword in section.

        :param section: section object returned from find_val

        Section is object found by previous find_val call.
        """
        if not isinstance(section, ConfigSection):
            raise TypeError("section must be ConfigSection")
        return self.find_val(section.config, key, section.start+1, section.end)

    def find_values(self, section, key):
        """Find key in section and list variable parameters.

        :param key: Name to statement to find
        :returns: List of all found values in form of ConfigSection. First is key itself.

        Returns all sections of keyname. They can be mix of "quoted strings", {nested blocks}
        or just bare keywords. First key is section of key itself, final section includes ';'.
        Makes it possible to comment out whole section including terminal character.
        """

        if isinstance(section, ConfigFile):
            cfg = section
            index = 0
            end_index = len(cfg.buffer)
        elif isinstance(section, ConfigSection):
            cfg = section.config
            index = section.start+1
            end_index = section.end
            if end_index > index:
                end_index -= 1
        else:
            raise TypeError('Unexpected type')

        if key is None:
            v = self.find_next_key(cfg, index, end_index)
        else:
            key_start = self.find_key(cfg.buffer, key, index, end_index)
            key_end = key_start+len(key)-1
            if key_start < 0 or key_end >= end_index:
                return None
            # First value is always just keyword
            v = ConfigSection(cfg, key, key_start, key_end)

        values = []
        while isinstance(v, ConfigSection):
            values.append(v)
            if v.value() == self.CHAR_DELIM:
                break
            v = self.find_next_val(cfg, key, v.end+1, end_index, end_report=True)
        return values

    def find(self, key_string, cfg=None, delimiter='.'):
        """Helper searching for values under requested sections.

        Search for statement under some sections. It is inspired by xpath style paths,
        but searches section in bind configuration.

        :param key_string: keywords delimited by dots. For example options.dnssec-lookaside
        :type key_string: str
        :param cfg: Search only in given config file
        :type cfg: ConfigFile
        :returns: list of ConfigVariableSection
        """
        keys = key_string.split(delimiter)
        if cfg is not None:
            return self._find_values_simple(cfg.root_section(), keys)

        items = []
        for cfgs in self.FILES_TO_CHECK:
            items.extend(self._find_values_simple(cfgs.root_section(), keys))
        return items

    def is_terminal(self, section):
        """.Returns true when section is final character of one statement."""
        return section.value() in self.CHAR_DELIM

    def _variable_section(self, vl, parent=None, offset=1):
        """Create ConfigVariableSection with a name and optionally class.

        Intended for view and zone in bind.
        :returns: ConfigVariableSection
        """
        vname = self._list_value(vl, 1).invalue()
        vclass = None
        v = self._list_value(vl, 2)
        if v.type() != ConfigSection.TYPE_BLOCK and self._list_value(vl, 2):
            vclass = v.value()
        return ConfigVariableSection(vl, vname, vclass, parent)

    def _list_value(self, vl, i):
        n = 0
        for v in vl:
            if v.type() != ConfigSection.TYPE_IGNORED:
                if n == i:
                    return v
                n += 1
        raise IndexError

    def _find_values_simple(self, section, keys):
        found_values = []
        sect = section.copy()

        while sect is not None:
            vl = self.find_values(sect, keys[0])
            if vl is None:
                break
            if len(keys) <= 1:
                variable = self._variable_section(vl, section)
                found_values.append(variable)
                sect.start = variable.end+1
            else:
                for v in vl:
                    if v.type() == ConfigSection.TYPE_BLOCK:
                        vl2 = self._find_values_simple(v, keys[1:])
                        if vl2 is not None:
                            found_values.extend(vl2)
                sect.start = vl[-1].end+1

        return found_values

    def walk(self, section, callbacks, state=None, parent=None, start=0):
        """Walk over section also with nested blocks.

        :param section: Section to iterate, usually ConfigFile.root_section()
        :param callbacks: Set of callbacks with name: f(section, state) parameters, indexed by statement name
        :param start: Offset from beginning of section

        Call specified actions specified in callbacks, which can react on desired statements.
        Pass state and matching section to callback.
        """
        if start == 0 and section.type() == ConfigSection.TYPE_BLOCK:
            start = 1
        it = IscVarIterator(self, section, True, start=section.start+start)
        for statement in it:
            try:
                name = statement.var(0).value()
                if name in callbacks:
                    f = callbacks[name]
                    f(statement, state)
            except IndexError:
                pass
            for child in statement.values:
                if child.type() == ConfigSection.TYPE_BLOCK:
                    self.walk(child, callbacks, state, parent=statement)
        return state

    #
    # CONFIGURATION fixes PART - END
    #

    def is_file_loaded(self, path=""):
        """
        Checks if the file with a given 'path' is already loaded in FILES_TO_CHECK.
        """
        for f in self.FILES_TO_CHECK:
            if f.path == path:
                return True
        return False

    def new_config(self, path, parent=None):
        config = ConfigFile(path)
        self.FILES_TO_CHECK.append(config)
        return config

    def on_include_error(self, e):
        """Handle IO errors on file reading.

        Override to create custom error handling."""
        raise e

    def load_included_files(self):
        """Add included list to parser.

        Finds the configuration files that are included in some configuration
        file, reads it, closes and adds into the FILES_TO_CHECK list.
        """
        # TODO: use parser instead of regexp
        pattern = re.compile(r'include\s*"(.+?)"\s*;')
        # find includes in all files
        for ch_file in self.FILES_TO_CHECK:
            nocomments = self._remove_comments(ch_file.buffer)
            includes = re.findall(pattern, nocomments)
            for include in includes:
                # don't include already loaded files -> prevent loops
                if self.is_file_loaded(include):
                    continue
                try:
                    self.new_config(include)
                except IOError as e:
                    self.on_include_error(ConfigParseError(e, include))

    def load_main_config(self):
        """Loads main CONFIG_FILE."""
        try:
            self.new_config(self.CONFIG_FILE)
        except IOError as e:
            raise ConfigParseError(e)

    def load_config(self, path=None):
        """Loads main config file with all included files."""
        if path is not None:
            self.CONFIG_FILE = path
        self.load_main_config()
        self.load_included_files()
    pass


if __name__ == '__main__':
    """Run parser to default path or path in the first argument.

    Additional parameters are statements or blocks to print.
    Defaults to options and zone.
    """

    from sys import argv

    def print_cb(section, state):
        print(section)

    cfgpath = IscConfigParser.CONFIG_FILE
    if len(argv) > 1:
        cfgpath = argv[1]
    if len(argv) > 2:
        cb = {}
        for key in argv[2:]:
            cb[key] = print_cb
    else:
        cb = {'options': print_cb, 'zone': print_cb}

    parser = IscConfigParser(cfgpath)
    for section in parser.FILES_TO_CHECK:
        print("# Walking file '{}'".format(section.path))
        parser.walk(section.root_section(), cb)



from leapp.models import fields, Model, RPM
from leapp.topics import SystemInfoTopic
from leapp.utils.deprecation import deprecated


@deprecated(since='2023-08-03', message='The model has been deprecated in favour of InstalledTargetKernelInfo.')
class InstalledTargetKernelVersion(Model):
    """
    This message is used to propagate the version of the kernel that has been installed during the upgrade process.

    The version value is mainly used for boot loader entry manipulations, to know which boot entry to manipulate.
    """
    topic = SystemInfoTopic
    version = fields.String()


class KernelInfo(Model):
    """
    Information about the booted kernel.
    """
    topic = SystemInfoTopic

    pkg = fields.Model(RPM)
    """ Package providing the booted kernel. """

    uname_r = fields.String()
    """``uname -r`` of the booted kernel."""

    type = fields.StringEnum(['ordinary', 'realtime'], default='ordinary')
    # @FixMe(mhecko): I want to use kernel_lib.KernelType here, but I cannot import any library code (yet).
    # #               Figure out how to do it.


class InstalledTargetKernelInfo(Model):
    """Information about the installed target kernel."""
    topic = SystemInfoTopic

    pkg_nevra = fields.String()
    """Name, epoch, version, release, arch of the target kernel package."""

    uname_r = fields.String()
    """Kernel release of the target kernel."""

    kernel_img_path = fields.String()
    """Path to the vmlinuz kernel image stored in ``/boot``."""

    initramfs_path = fields.String()
    """Path to the initramfs image stored in ``/boot``."""



from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.config import architecture, version
from leapp.libraries.stdlib import api
from leapp.models import MemoryInfo

min_req_memory = {
    architecture.ARCH_X86_64: 1572864,  # 1.5G
    architecture.ARCH_ARM64: 1572864,  # 1.5G
    architecture.ARCH_PPC64LE: 3145728,  # 3G
    architecture.ARCH_S390X: 1572864,  # 1.5G
}


def _check_memory(mem_info):
    msg = {}

    for arch, min_req in iter(min_req_memory.items()):
        if architecture.matches_architecture(arch):
            is_ok = mem_info.mem_total >= min_req
            msg = {} if is_ok else {'detected': mem_info.mem_total,
                                    'minimal_req': min_req}

    return msg


def process():
    memoryinfo = next(api.consume(MemoryInfo), None)
    if memoryinfo is None:
        raise StopActorExecutionError(message="Missing information about Memory.")

    minimum_req_error = _check_memory(memoryinfo)

    if minimum_req_error:
        title = 'Minimum memory requirements for RHEL {} are not met'.format(version.get_target_major_version())
        summary = 'Memory detected: {} MiB, required: {} MiB'.format(
            int(minimum_req_error['detected'] / 1024),  # noqa: W1619; pylint: disable=old-division
            int(minimum_req_error['minimal_req'] / 1024),  # noqa: W1619; pylint: disable=old-division
        )
        reporting.create_report([
                          reporting.Title(title),
                          reporting.Summary(summary),
                          reporting.Severity(reporting.Severity.HIGH),
                          reporting.Groups([reporting.Groups.SANITY, reporting.Groups.INHIBITOR]),
                          reporting.ExternalLink(
                            url='https://access.redhat.com/articles/rhel-limits',
                            title='Red Hat Enterprise Linux Technology Capabilities and Limits'
                          ),
                      ])



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class GrubInfo(Model):
    """
    Information about Grub
    """
    topic = SystemFactsTopic

    # NOTE: @deprecated is not supported on fields
    # @deprecated(since='2023-06-23', message='This field has been replaced by orig_devices')
    orig_device_name = fields.Nullable(fields.String())
    """
    Original name of the block device where Grub is located.

    The name is persistent during the boot of the system so it's safe to be used during
    preupgrade phases. However the name could be different after the reboot, so
    it's recommended to use `leapp.libraries.common.grub.get_grub_device()` anywhere
    else.
    """

    orig_devices = fields.List(fields.String(), default=[])
    """
    Original names of the block devices where Grub is located.

    The names are persistent during the boot of the system so it's safe to be used during
    preupgrade phases. However the names could be different after the reboot, so
    it's recommended to use `leapp.libraries.common.grub.get_grub_devices()` everywhere
    else.
    """



import os

from leapp.libraries.actor import sssd_cache_files


def glob_sssd_files(path_spec):
    if path_spec.endswith('*.ldb'):
        path_spec = path_spec[:-5]
        return list([os.path.join(path_spec, 'removed_file1.ldb'), os.path.join(path_spec, 'removed_file2.ldb')])
    return []


def test_remove_sssd_cache_files(monkeypatch):
    monkeypatch.setattr(
        sssd_cache_files.glob,
        'glob',
        glob_sssd_files)
    removed = []
    monkeypatch.setattr(sssd_cache_files.os, 'remove', removed.append)
    sssd_cache_files.remove_sssd_cache_files(None)
    assert not removed
    sssd_cache_files.remove_sssd_cache_files(True)
    assert removed == ['/var/lib/sss/db/removed_file1.ldb', '/var/lib/sss/db/removed_file2.ldb']



from leapp.actors import Actor
from leapp.libraries.actor.sssd_cache_files import remove_sssd_cache_files
from leapp.models import SSSDConfig
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class SSSDHandleCacheFiles(Actor):
    """
    If an SSSD configuration has been detected, special handling for SSSD cache files will be scheduled.

    After the upgrade from EL7 to EL8, SSSD isn't able to start anymore because the cache files have a newer
    format in RHEL7.9 than in RHEL8. This is a workaround for this issue.
    """

    name = 'sssd_handle_cache_files'
    consumes = (SSSDConfig,)
    produces = ()
    tags = (IPUWorkflowTag, ApplicationsPhaseTag)

    def process(self):
        remove_sssd_cache_files(
            next(self.consume(SSSDConfig), None)
        )



import os

from leapp.libraries.common import repofileutils
from leapp.libraries.stdlib import api
from leapp.models import CustomTargetRepository, CustomTargetRepositoryFile

CUSTOM_REPO_PATH = "/etc/leapp/files/leapp_upgrade_repositories.repo"


def process():
    """
    Produce CustomTargetRepository msgs for the custom repo file if the file
    exists.

    The CustomTargetRepository msg is produced for every repository inside
    the <CUSTOM_REPO_PATH> file.
    """
    if not os.path.isfile(CUSTOM_REPO_PATH):
        api.current_logger().debug(
                "The {} file doesn't exist. Nothing to do."
                .format(CUSTOM_REPO_PATH))
        return
    api.current_logger().info("The {} file exists.".format(CUSTOM_REPO_PATH))
    repofile = repofileutils.parse_repofile(CUSTOM_REPO_PATH)
    if not repofile.data:
        return
    api.produce(CustomTargetRepositoryFile(file=CUSTOM_REPO_PATH))
    for repo in repofile.data:
        api.produce(CustomTargetRepository(
            repoid=repo.repoid,
            name=repo.name,
            baseurl=repo.baseurl,
            enabled=repo.enabled,
        ))



from leapp.actors import Actor
from leapp.libraries.actor import checkfstabxfsoptions
from leapp.models import StorageInfo
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckFstabXFSOptions(Actor):
    """
    Check the FSTAB file for the deprecated / removed XFS mount options.

    Some mount options for XFS have been deprecated on RHEL 7 and already
    removed on RHEL 8. If any such an option is present in the FSTAB,
    it's impossible to boot the RHEL 8 system without the manual update of the
    file.

    Check whether any of these options are present in the FSTAB file
    and inhibit the upgrade in such a case.
    """

    name = 'checkfstabxfsoptions'
    consumes = (StorageInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        checkfstabxfsoptions.process()



from leapp.actors import Actor
from leapp.libraries.actor import setsystemdservicesstate
from leapp.models import SystemdServicesTasks
from leapp.tags import FinalizationPhaseTag, IPUWorkflowTag


class SetSystemdServicesState(Actor):
    """
    According to input messages sets systemd services states on the target system
    """

    name = 'set_systemd_services_state'
    consumes = (SystemdServicesTasks,)
    produces = ()
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def process(self):
        setsystemdservicesstate.process()



from leapp.libraries.common import config, mounting, rhsm
from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import UsedTargetRepositories


def set_rhsm_release():
    """Set the RHSM release to the target RHEL minor version."""
    if rhsm.skip_rhsm():
        api.current_logger().debug('Skipping setting the RHSM release due to --no-rhsm or environment variables.')
        return

    if config.get_product_type('target') == 'beta':
        api.current_logger().debug('Skipping setting the RHSM release as target product is set to beta')
        return
    target_version = api.current_actor().configuration.version.target
    try:
        rhsm.set_release(mounting.NotIsolatedActions(base_dir='/'), target_version)
    except CalledProcessError as err:
        api.current_logger().warning('Unable to set the {0} release through subscription-manager. When using dnf,'
                                     ' content of the latest RHEL {1} minor version will be downloaded.\n{2}'
                                     .format(target_version, get_target_major_version(), str(err)))


def enable_rhsm_repos():
    """
    Try enabling all the target RHEL repositories that have been used for the upgrade transaction.

    In case of custom repositories, the subscription-manager reports an error that it doesn't know them, but it enables
    the known repositories.
    """
    if rhsm.skip_rhsm():
        api.current_logger().debug('Skipping enabling repositories through subscription-manager due to --no-rhsm'
                                   ' or environment variables.')
        return
    try:
        run(get_submgr_cmd(get_repos_to_enable()))
    except CalledProcessError as err:
        api.current_logger().warning('The subscription-manager could not enable some repositories.\n'
                                     'It is expected behavior in case of custom repositories unknown to'
                                     ' the subscription-manager - these need to be enabled manually.\n{0}'
                                     .format(str(err)))


def get_submgr_cmd(repos_to_enable):
    submgr_cmd = ['subscription-manager', 'repos']
    for repoid in repos_to_enable:
        submgr_cmd += ['--enable', repoid]
    return submgr_cmd


def get_repos_to_enable():
    """
    Return set of repositories used during the upgrade transaction.

    This set may include repos unknown to subscription-manager - notable those added as custom using
    CustomTargetRepository model.
    """
    used_repos_msg = next(api.consume(UsedTargetRepositories), None)
    return {repo.repoid for repo in used_repos_msg.repos}



import pytest

from leapp import reporting
from leapp.libraries.actor import ziplcheckbootentries
from leapp.libraries.actor.ziplcheckbootentries import (
    extract_kernel_version,
    inhibit_if_entries_share_kernel_version,
    inhibit_if_invalid_zipl_configuration,
    inhibit_if_multiple_zipl_rescue_entries_present
)
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import BootEntry, SourceBootLoaderConfiguration
from leapp.snactor.fixture import current_actor_context
from leapp.utils.report import is_inhibitor


def test_inhibition_multiple_rescue_entries_present(monkeypatch):
    """Tests whether the upgrade process is inhibited when multiple rescue boot entries are present."""
    mocked_report = create_report_mocked()
    monkeypatch.setattr(architecture, 'matches_architecture', lambda dummy: True)
    monkeypatch.setattr(reporting, 'create_report', mocked_report)

    boot_entries = [
        BootEntry(title='entry_1', kernel_image="img"),
        BootEntry(title='entry_1_Rescue', kernel_image="img_Rescue"),
        BootEntry(title='entry_2', kernel_image="img"),
        BootEntry(title='entry_2_rescue-ver2.3', kernel_image="img_rescue"),
    ]

    inhibit_if_multiple_zipl_rescue_entries_present(SourceBootLoaderConfiguration(entries=boot_entries))

    assert mocked_report.called, 'Report should be created when multiple rescue entries are present.'

    fail_description = 'The correct rescue entries are not present in the report summary.'
    report_summary = mocked_report.report_fields['summary']
    for expected_rescue_entry in ['entry_1_Rescue', 'entry_2_rescue-ver2.3']:
        assert expected_rescue_entry in report_summary, fail_description

    fail_description = 'Upgrade should be inhibited on multiple rescue entries.'
    assert is_inhibitor(mocked_report.report_fields), fail_description


def test_inhibition_multiple_rescue_entries_not_present(monkeypatch):
    """Tests whether the upgrade process is not inhibited when multiple rescue boot entries are not present."""
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    boot_entries = [
        BootEntry(title='entry_1', kernel_image="img"),
        BootEntry(title='entry_2', kernel_image="img"),
        BootEntry(title='entry_2_rescue-ver2.3', kernel_image="img_rescue"),
    ]

    inhibit_if_multiple_zipl_rescue_entries_present(SourceBootLoaderConfiguration(entries=boot_entries))

    assert not reporting.create_report.called, 'Report was created, even if multiple rescue entries were not present.'


def test_inhibition_when_entries_do_not_share_kernel_image(monkeypatch):
    """Tests whether the IPU is not inhibited when there are no kernel images shared between boot entries."""
    entries = [
        BootEntry(title='Linux#0', kernel_image='/boot/vmlinuz-4.17.0-240.1.1.el8_3.x86_64'),
        BootEntry(title='Linux#1', kernel_image='/boot/vmlinuz-4.18.0-240.1.1.el8_3.x86_64')
    ]

    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    inhibit_if_entries_share_kernel_version(SourceBootLoaderConfiguration(entries=entries))
    assert not reporting.create_report.called


@pytest.mark.parametrize(
    ('boot_entries',),
    [([BootEntry(title='Linux0', kernel_image='/boot/vmlinuz-4.18.0-240.1.1.el8_3.x86_64'),
       BootEntry(title='Linux1', kernel_image='/boot/4.18.0-240.1.1.el8_3.x86_64')],),
     ([BootEntry(title='Linux0', kernel_image='/boot/vmlinuz-4.18.0-240.1.1.el8_3.x86_64'),
       BootEntry(title='Linux1', kernel_image='/boot/vmlinuz-4.18.0-240.1.1.el8_3.x86_64')],)])
def test_inhibit_when_entries_share_kernel_image(monkeypatch, boot_entries):
    """Tests whether the IPU gets inhibited when there are kernel images shared between boot entries."""

    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    inhibit_if_entries_share_kernel_version(SourceBootLoaderConfiguration(entries=boot_entries))

    assert reporting.create_report.called
    assert is_inhibitor(reporting.create_report.report_fields)

    report_summary = reporting.create_report.report_fields['summary']
    assert '- 4.18.0-240.1.1.el8_3.x86_64 (found in entries: "Linux0", "Linux1")' in report_summary


@pytest.mark.parametrize(
    ('boot_entries',),
    [([BootEntry(title='Linux', kernel_image='/boot/vmlinuz-4.18.0-240.1.1.el8_3.x86_64'),
       BootEntry(title='Linux-rescue', kernel_image='/boot/vmlinuz-rescue-4.18.0-240.1.1.el8_3.x86_64')],),
     ([BootEntry(title='Linux0-rescue', kernel_image='/boot/vmlinuz-rescue-4.18.0-240.1.1.el8_3.x86_64'),
       BootEntry(title='Linux1-rescue', kernel_image='/boot/vmlinuz-rescue-4.18.0-240.1.1.el8_3.x86_64')],)])
def test_inhibition_when_rescue_entries_share_kernel(monkeypatch, boot_entries):
    """
    Tests whether the IPU is not inhibited when there are kernel images with the same version shared between rescue
    boot entries.
    """
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    inhibit_if_entries_share_kernel_version(SourceBootLoaderConfiguration(entries=boot_entries))
    assert not reporting.create_report.called


@pytest.mark.parametrize(('arch',), [(arch,) for arch in architecture.ARCH_SUPPORTED])
def test_checks_performed_only_on_s390x_arch(arch, monkeypatch):
    """Tests whether the actor doesn't perform different architectures than s390x."""
    should_perform = False
    if arch == architecture.ARCH_S390X:  # Rescue entries should be checked only on s390x.
        should_perform = True

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch=arch))
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    boot_entries = [BootEntry(title='rescue0', kernel_image='/boot/vmlinuz-rescue-4.18.0-240.1.1.el8_3.x86_64'),
                    BootEntry(title='rescue1', kernel_image='/boot/vmlinuz-rescue-4.19.0-240.1.1.el8_3.x86_64')]

    inhibit_if_invalid_zipl_configuration(SourceBootLoaderConfiguration(entries=boot_entries))

    fail_description = 'Rescue entries should not be checked on non s390x architecture.'
    if should_perform:
        fail_description = 'No report was created when running on s390x and multiple rescue entries were used.'
    assert bool(reporting.create_report.called) == should_perform, fail_description

    if should_perform:
        inhibitor_description = 'contains multiple rescue boot entries'
        assert inhibitor_description in reporting.create_report.report_fields['summary']

    boot_entries = [BootEntry(title='Linux1', kernel_image='/boot/vmlinuz-4.18.0-240.1.1.el8_3.x86_64'),
                    BootEntry(title='Linux2', kernel_image='/boot/vmlinuz-4.18.0-240.1.1.el8_3.x86_64')]

    inhibit_if_invalid_zipl_configuration(SourceBootLoaderConfiguration(entries=boot_entries))

    fail_description = 'Check for boot entries with the same kernel version should not be performed on non s390x arch.'
    if should_perform:
        fail_description = ('No report was created when running on s390x and boot entries'
                            'with the same kernel version are present')
    assert bool(reporting.create_report.called) == should_perform, fail_description
    if should_perform:
        inhibitor_description = 'contains boot entries sharing the same kernel version'
        assert inhibitor_description in reporting.create_report.report_fields['summary']


def test_extract_kernel_version():
    # Manually generated via experimentation with the zipl-switch-to-blscfg
    versions_from_img_paths = [
        ('/boot/vmlinuz-4.18.0-240.1.1.el8_3.x86_64', '4.18.0-240.1.1.el8_3.x86_64'),
        ('/boot/4.18.0-240.1.1.el8_3.x86_64', '4.18.0-240.1.1.el8_3.x86_64'),
        ('/boot/patched-4.18.0-240.1.1.el8_3.x86_64', 'patched-4.18.0-240.1.1.el8_3.x86_64'),
        ('patched-4.18.0-240.1.1.el8_3.x86_64', 'patched-4.18.0-240.1.1.el8_3.x86_64'),
    ]

    for path, version in versions_from_img_paths:
        assert extract_kernel_version(path) == version



from leapp.actors import Actor
from leapp.libraries.actor.nisscan import NISScanLibrary
from leapp.models import NISConfig
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class NISScanner(Actor):
    """
    Collect information about the NIS packages configuration.
    """

    name = 'nis_scanner'
    consumes = ()
    produces = (NISConfig,)
    tags = (FactsPhaseTag, IPUWorkflowTag,)

    def process(self):
        NISScanLibrary().process()



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class HybridImage(Model):
    """
    Model used for instructing Leapp to convert "grubenv" symlink
    into a regular file in case of hybrid (BIOS/EFI) images using BIOS
    on Azure.
    """
    topic = SystemFactsTopic
    detected = fields.Boolean(default=False)



import pytest

from leapp import reporting
from leapp.libraries.actor import checksystemdbrokensymlinks
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import SystemdBrokenSymlinksSource, SystemdServiceFile, SystemdServicesInfoSource


def test_report_broken_symlinks(monkeypatch):

    symlinks = [
        '/etc/systemd/system/multi-user.target.wants/hello.service',
        '/etc/systemd/system/multi-user.target.wants/world.service',
    ]

    created_reports = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', created_reports)

    checksystemdbrokensymlinks._report_broken_symlinks(symlinks)

    assert created_reports.called
    assert all([s in created_reports.report_fields['summary'] for s in symlinks])


def test_report_enabled_services_broken_symlinks(monkeypatch):
    symlinks = [
        '/etc/systemd/system/multi-user.target.wants/foo.service',
        '/etc/systemd/system/multi-user.target.wants/bar.service',
    ]

    created_reports = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', created_reports)

    checksystemdbrokensymlinks._report_enabled_services_broken_symlinks(symlinks)

    assert created_reports.called
    assert all([s in created_reports.report_fields['summary'] for s in symlinks])


class ReportBrokenSymlinks(object):
    def __init__(self):
        self.symlinks = []

    def __call__(self, unit, *args, **kwargs):
        self.symlinks.append(unit)
        return {}


def test_broken_symlinks_reported(monkeypatch):
    broken_symlinks = SystemdBrokenSymlinksSource(broken_symlinks=[
        '/etc/systemd/system/multi-user.target.wants/foo.service',
        '/etc/systemd/system/multi-user.target.wants/bar.service',
        '/etc/systemd/system/multi-user.target.wants/hello.service',
        '/etc/systemd/system/multi-user.target.wants/world.service',
    ])
    systemd_services = SystemdServicesInfoSource(service_files=[
        SystemdServiceFile(name='foo.service', state='enabled'),
        SystemdServiceFile(name='bar.service', state='enabled'),
        SystemdServiceFile(name='hello.service', state='disabled'),
    ])
    broken = []
    enabled_broken = []

    def _report_broken_symlinks_mocked(symlinks):
        broken.extend(symlinks)

    def _report_enabled_services_broken_symlinks_mocked(symlinks):
        enabled_broken.extend(symlinks)

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[broken_symlinks, systemd_services]))
    monkeypatch.setattr(checksystemdbrokensymlinks, '_report_broken_symlinks', _report_broken_symlinks_mocked)
    monkeypatch.setattr(
        checksystemdbrokensymlinks,
        '_report_enabled_services_broken_symlinks',
        _report_enabled_services_broken_symlinks_mocked
    )

    checksystemdbrokensymlinks.process()

    assert broken == [
        '/etc/systemd/system/multi-user.target.wants/hello.service',
        '/etc/systemd/system/multi-user.target.wants/world.service',
    ]

    assert enabled_broken == [
        '/etc/systemd/system/multi-user.target.wants/foo.service',
        '/etc/systemd/system/multi-user.target.wants/bar.service',
    ]



from leapp.libraries.common import systemd
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import SystemdBrokenSymlinksTarget, SystemdServicesInfoTarget, SystemdServicesPresetInfoTarget


def scan_broken_symlinks():
    try:
        broken_symlinks = systemd.get_broken_symlinks()
    except (OSError, CalledProcessError):
        return
    api.produce(SystemdBrokenSymlinksTarget(broken_symlinks=broken_symlinks))


def scan_service_files():
    try:
        services_files = systemd.get_service_files()
    except CalledProcessError:
        return None
    api.produce(SystemdServicesInfoTarget(service_files=services_files))
    return services_files


def scan_preset_files(services_files):
    if services_files is None:
        return
    try:
        presets = systemd.get_system_service_preset_files(services_files, ignore_invalid_entries=True)
    except (OSError, CalledProcessError):
        return
    api.produce(SystemdServicesPresetInfoTarget(presets=presets))


def scan():
    # Errors are logged inside the systemd library, no need to log them here again.
    scan_broken_symlinks()
    services_files = scan_service_files()
    scan_preset_files(services_files)



import os

import pytest

from leapp import reporting
from leapp.libraries.actor import check_target_iso
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import CustomTargetRepository, FstabEntry, StorageInfo, TargetOSInstallationImage
from leapp.utils.report import is_inhibitor


@pytest.mark.parametrize('mount_successful', (True, False))
def test_inhibit_on_iso_mount_failure(monkeypatch, mount_successful):
    create_report_mock = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', create_report_mock)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())

    target_iso_msg = TargetOSInstallationImage(path='',
                                               mountpoint='',
                                               repositories=[],
                                               was_mounted_successfully=mount_successful)

    check_target_iso.inhibit_if_failed_to_mount_iso(target_iso_msg)

    expected_report_count = 0 if mount_successful else 1
    assert create_report_mock.called == expected_report_count
    if not mount_successful:
        assert is_inhibitor(create_report_mock.reports[0])


@pytest.mark.parametrize(('detected_iso_rhel_ver', 'required_target_ver', 'should_inhibit'),
                         (('8.6', '8.6', False), ('7.9', '8.6', True), ('8.5', '8.6', False), ('', '8.6', True)))
def test_inhibit_on_detected_rhel_version(monkeypatch, detected_iso_rhel_ver, required_target_ver, should_inhibit):
    create_report_mock = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', create_report_mock)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(dst_ver=required_target_ver))

    target_iso_msg = TargetOSInstallationImage(path='',
                                               mountpoint='',
                                               repositories=[],
                                               rhel_version=detected_iso_rhel_ver,
                                               was_mounted_successfully=True)

    check_target_iso.inhibit_if_wrong_iso_rhel_version(target_iso_msg)

    expected_report_count = 1 if should_inhibit else 0
    assert create_report_mock.called == expected_report_count
    if should_inhibit:
        assert is_inhibitor(create_report_mock.reports[0])


@pytest.mark.parametrize(('iso_repoids', 'should_inhibit'),
                         ((('BaseOS', 'AppStream'), False), (('BaseOS',), True), (('AppStream',), True), ((), True)))
def test_inhibit_on_invalid_rhel_version(monkeypatch, iso_repoids, should_inhibit):
    create_report_mock = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', create_report_mock)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())

    iso_repositories = [CustomTargetRepository(repoid=repoid, baseurl='', name='') for repoid in iso_repoids]

    target_iso_msg = TargetOSInstallationImage(path='',
                                               mountpoint='',
                                               repositories=iso_repositories,
                                               was_mounted_successfully=True)

    check_target_iso.inihibit_if_iso_does_not_contain_basic_repositories(target_iso_msg)

    expected_report_count = 1 if should_inhibit else 0
    assert create_report_mock.called == expected_report_count
    if should_inhibit:
        assert is_inhibitor(create_report_mock.reports[0])


def test_inhibit_on_nonexistent_iso(monkeypatch):
    iso_path = '/nonexistent/iso'
    create_report_mock = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', create_report_mock)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())

    def mocked_os_path_exists(path):
        assert path == iso_path, 'The actor should check only the path to ISO for existence.'
        return False

    monkeypatch.setattr(os.path, 'exists', mocked_os_path_exists)

    target_iso_msg = TargetOSInstallationImage(path=iso_path,
                                               mountpoint='',
                                               repositories=[],
                                               was_mounted_successfully=True)

    check_target_iso.inhibit_if_not_valid_iso_file(target_iso_msg)

    assert create_report_mock.called == 1
    assert is_inhibitor(create_report_mock.reports[0])


@pytest.mark.parametrize(('filetype', 'should_inhibit'),
                         (('{path}: text/plain; charset=us-ascii', True),
                          ('{path}: application/x-iso9660-image; charset=binary', False)))
def test_inhibit_on_path_not_pointing_to_iso(monkeypatch, filetype, should_inhibit):
    iso_path = '/path/not-an-iso'
    create_report_mock = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', create_report_mock)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())

    def mocked_os_path_exists(path):
        assert path == iso_path, 'The actor should check only the path to ISO for existence.'
        return True

    def mocked_run(cmd, *args, **kwargs):
        assert cmd[0] == 'file', 'The actor should only use `file` cmd when checking for file type.'
        return {'stdout': filetype.format(path=iso_path)}

    monkeypatch.setattr(os.path, 'exists', mocked_os_path_exists)
    monkeypatch.setattr(check_target_iso, 'run', mocked_run)

    target_iso_msg = TargetOSInstallationImage(path=iso_path, mountpoint='', repositories=[])

    check_target_iso.inhibit_if_not_valid_iso_file(target_iso_msg)

    if should_inhibit:
        assert create_report_mock.called == 1
        assert is_inhibitor(create_report_mock.reports[0])
    else:
        assert create_report_mock.called == 0


@pytest.mark.parametrize('is_persistently_mounted', (False, True))
def test_inhibition_when_iso_not_on_persistent_partition(monkeypatch, is_persistently_mounted):
    path_mountpoint = '/d0/d1'
    iso_path = '/d0/d1/d2/d3/iso'
    create_report_mock = create_report_mocked()
    monkeypatch.setattr(reporting, 'create_report', create_report_mock)

    def os_path_ismount_mocked(path):
        if path == path_mountpoint:
            return True
        if path == '/':  # / Should be a mountpoint on every system
            return True
        return False

    monkeypatch.setattr(os.path, 'ismount', os_path_ismount_mocked)

    fstab_mountpoint = path_mountpoint if is_persistently_mounted else '/some/other/mountpoint'
    fstab_entry = FstabEntry(fs_spec='/dev/sta2', fs_file=fstab_mountpoint,
                             fs_vfstype='', fs_mntops='', fs_freq='', fs_passno='')
    storage_info_msg = StorageInfo(fstab=[fstab_entry])

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[storage_info_msg]))

    target_iso_msg = TargetOSInstallationImage(path=iso_path, mountpoint='', repositories=[])
    check_target_iso.inhibit_if_iso_not_located_on_persistent_partition(target_iso_msg)

    if is_persistently_mounted:
        assert not create_report_mock.called
    else:
        assert create_report_mock.called == 1
        assert is_inhibitor(create_report_mock.reports[0])


def test_actor_does_not_perform_when_iso_not_used(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())

    check_target_iso.perform_target_iso_checks()

    assert not reporting.create_report.called



import pytest

from leapp.libraries.actor import removeobsoleterpmgpgkeys
from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.common.rpms import has_package
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import DNFWorkaround, InstalledRPM, RPM


def _get_test_installedrpm():
    return InstalledRPM(
        items=[
            RPM(
                name='gpg-pubkey',
                version='d4082792',
                release='5b32db75',
                epoch='0',
                packager='Red Hat, Inc. (auxiliary key 2) <security@redhat.com>',
                arch='noarch',
                pgpsig=''
            ),
            RPM(
                name='gpg-pubkey',
                version='2fa658e0',
                release='45700c69',
                epoch='0',
                packager='Red Hat, Inc. (auxiliary key) <security@redhat.com>',
                arch='noarch',
                pgpsig=''
            ),
            RPM(
                name='gpg-pubkey',
                version='12345678',
                release='abcdefgh',
                epoch='0',
                packager='made up',
                arch='noarch',
                pgpsig=''
            ),
        ]
    )


@pytest.mark.parametrize(
    "version, expected",
    [
        (9, ["gpg-pubkey-d4082792-5b32db75", "gpg-pubkey-2fa658e0-45700c69"]),
        (8, ["gpg-pubkey-2fa658e0-45700c69"])
    ]
)
def test_get_obsolete_keys(monkeypatch, version, expected):
    def get_target_major_version_mocked():
        return version

    monkeypatch.setattr(
        removeobsoleterpmgpgkeys,
        "get_target_major_version",
        get_target_major_version_mocked,
    )

    monkeypatch.setattr(
        api,
        "current_actor",
        CurrentActorMocked(
            msgs=[_get_test_installedrpm()]
        ),
    )

    keys = removeobsoleterpmgpgkeys._get_obsolete_keys()
    assert set(keys) == set(expected)


@pytest.mark.parametrize(
    "keys, should_register",
    [
        (["gpg-pubkey-d4082792-5b32db75"], True),
        ([], False)
    ]
)
def test_workaround_should_register(monkeypatch, keys, should_register):
    def get_obsolete_keys_mocked():
        return keys

    monkeypatch.setattr(
        removeobsoleterpmgpgkeys,
        '_get_obsolete_keys',
        get_obsolete_keys_mocked
    )
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(api, "current_actor", CurrentActorMocked())

    removeobsoleterpmgpgkeys.process()
    assert api.produce.called == should_register



import errno
import re

from leapp.libraries.stdlib import api
from leapp.models import DaemonList, TcpWrappersFacts

CONFIG_FILES = ['/etc/hosts.allow', '/etc/hosts.deny']


def _read_file(path):
    with open(path, 'r') as f:
        return f.read()


def _get_lines(content):
    content = re.sub(r'\\\n', '', content)
    return content.split('\n')


def _is_comment(line):
    return len(line) == 0 or line.isspace() or line.startswith('#')


def _get_daemon_list_in_line(line):
    daemon_list = line.split(':', 1)[0]
    daemon_list = re.split(',| |\t', daemon_list)
    daemon_list = [word for word in daemon_list if len(word) > 0]
    return DaemonList(value=daemon_list)


def _get_daemon_lists_in_file(path, read_func=_read_file):
    ret = []
    try:
        content = read_func(path)
    except IOError as e:
        if e.errno != errno.ENOENT:
            api.current_logger().warning('Failed to read %s: %s' % (path, e))
        return ret
    lines = [line for line in _get_lines(content) if not _is_comment(line)]
    for line in lines:
        ret.append(_get_daemon_list_in_line(line))
    return ret


def _get_daemon_lists(read_func=_read_file):
    daemon_lists = []
    for path in CONFIG_FILES:
        daemon_lists.extend(_get_daemon_lists_in_file(path, read_func=read_func))
    return daemon_lists


def get_tcp_wrappers_facts(read_func=_read_file):
    daemon_lists = _get_daemon_lists(read_func=read_func)
    return TcpWrappersFacts(daemon_lists=daemon_lists)



from leapp.actors import Actor
from leapp.libraries.actor import scansourcefiles
from leapp.models import TrackedFilesInfoSource
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanSourceFiles(Actor):
    """
    Scan files (explicitly specified) of the source system.

    If an actor require information about a file, like whether it's installed,
    modified, etc. It can be added to the list of files to be tracked, so no
    extra actor is required to be created to provide just that one information.

    The scan of all changed files tracked by RPMs is very expensive. So we rather
    provide this possibility to simplify the work for others.

    See lists defined in the private library.
    """
    # TODO(pstodulk): in some cases could be valuable to specify an rpm name
    # and provide information about all changed files instead. Both approaches
    # have a little bit different use-cases and expectations. In the second
    # case it would be good solution regarding track of leapp-repository
    # changed files.

    name = 'scan_source_files'
    consumes = ()
    produces = (TrackedFilesInfoSource,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        scansourcefiles.process()



from leapp import reporting
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM

# Summary for postgresql-server report
report_server_inst_summary = (
    'PostgreSQL server component will be upgraded. Since RHEL-9 includes'
    ' PostgreSQL server 13 by default, which is incompatible with 9.6, 10 and 12'
    ' included in RHEL-8, it is necessary to proceed with additional steps'
    ' for the complete upgrade of the PostgreSQL data.'
)

report_server_inst_hint = (
    'Back up your data before proceeding with the upgrade'
    ' and follow steps in the documentation section "Migrating to a RHEL 9 version of PostgreSQL"'
    ' after the upgrade.'
)

# Link URL for postgresql-server report
report_server_inst_link_url = 'https://access.redhat.com/articles/6654721'


def _report_server_installed():
    """
    Create report on postgresql-server package installation detection.

    Should remind user about present PostgreSQL server package
    installation, warn them about necessary additional steps, and
    redirect them to online documentation for the upgrade process.
    """
    reporting.create_report([
        reporting.Title('PostgreSQL (postgresql-server) has been detected on your system'),
        reporting.Summary(report_server_inst_summary),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.ExternalLink(title='Migrating to a RHEL 9 version of PostgreSQL',
                               url=report_server_inst_link_url),
        reporting.RelatedResource('package', 'postgresql-server'),
        reporting.Remediation(hint=report_server_inst_hint),
        ])


def report_installed_packages(_context=api):
    """
    Create reports according to detected PostgreSQL packages.

    Create the report if the postgresql-server rpm (RH signed) is installed.
    """
    has_server = has_package(DistributionSignedRPM, 'postgresql-server', context=_context)

    if has_server:
        # postgresql-server
        _report_server_installed()



from leapp.actors import Actor
from leapp.libraries.actor.quaggadaemons import process_daemons
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM, QuaggaToFrrFacts
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class QuaggaDaemons(Actor):
    """
    Active quagga daemons check.

    Checking for daemons that are currently running in the system.
    These should be enabled in /etc/frr/daemons later in the process.
    The tools will check for config files later on since these should stay in the system.
    """

    name = 'quagga_daemons'
    consumes = (DistributionSignedRPM,)
    produces = (QuaggaToFrrFacts,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(DistributionSignedRPM, 'quagga'):
            self.produce(process_daemons())



import pytest

from leapp.libraries.common import config
from leapp.models import FstabEntry, MountEntry, StorageInfo
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context
from leapp.utils.report import is_inhibitor


@pytest.mark.parametrize('nfs_fstype', ('nfs', 'nfs4'))
def test_actor_with_fstab_entry(current_actor_context, nfs_fstype, monkeypatch):
    monkeypatch.setattr(config, 'get_env', lambda x, y: y)
    with_fstab_entry = [FstabEntry(fs_spec="lithium:/mnt/data", fs_file="/mnt/data",
                                   fs_vfstype=nfs_fstype,
                                   fs_mntops="noauto,noatime,rsize=32768,wsize=32768",
                                   fs_freq="0", fs_passno="0")]
    current_actor_context.feed(StorageInfo(fstab=with_fstab_entry))
    current_actor_context.run()
    report_fields = current_actor_context.consume(Report)[0].report
    assert is_inhibitor(report_fields)


def test_actor_without_fstab_entry(current_actor_context, monkeypatch):
    monkeypatch.setattr(config, 'get_env', lambda x, y: y)
    without_fstab_entry = [FstabEntry(fs_spec="/dev/mapper/fedora-home", fs_file="/home",
                                      fs_vfstype="ext4",
                                      fs_mntops="defaults,x-systemd.device-timeout=0",
                                      fs_freq="1", fs_passno="2")]
    current_actor_context.feed(StorageInfo(fstab=without_fstab_entry))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_with_nfsd(current_actor_context, monkeypatch):
    monkeypatch.setattr(config, 'get_env', lambda x, y: y)
    with_nfsd = [MountEntry(name="nfsd", mount="/proc/fs/nfsd", tp="nfsd", options="rw,relatime")]
    current_actor_context.feed(StorageInfo(mount=with_nfsd))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


@pytest.mark.parametrize('nfs_fstype', ('nfs', 'nfs4'))
def test_actor_with_mount_share(current_actor_context, nfs_fstype, monkeypatch):
    monkeypatch.setattr(config, 'get_env', lambda x, y: y)
    with_mount_share = [MountEntry(name="nfs", mount="/mnt/data", tp=nfs_fstype,
                                   options="rw,nosuid,nodev,relatime,user_id=1000,group_id=1000")]
    current_actor_context.feed(StorageInfo(mount=with_mount_share))
    current_actor_context.run()
    report_fields = current_actor_context.consume(Report)[0].report
    assert is_inhibitor(report_fields)


def test_actor_without_mount_share(current_actor_context, monkeypatch):
    monkeypatch.setattr(config, 'get_env', lambda x, y: y)
    without_mount_share = [MountEntry(name="tmpfs", mount="/run/snapd/ns", tp="tmpfs",
                                      options="rw,nosuid,nodev,seclabel,mode=755")]
    current_actor_context.feed(StorageInfo(mount=without_mount_share))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_skipped_if_initram_network_enabled(current_actor_context, monkeypatch):
    """Check that previous inhibitors are not stopping the upgrade in case env var is set"""
    monkeypatch.setattr(config, 'get_env', lambda x, y: 'network-manager' if x == 'LEAPP_DEVEL_INITRAM_NETWORK' else y)
    with_mount_share = [MountEntry(name="nfs", mount="/mnt/data", tp='nfs',
                                   options="rw,nosuid,nodev,relatime,user_id=1000,group_id=1000")]
    with_fstab_entry = [FstabEntry(fs_spec="lithium:/mnt/data", fs_file="/mnt/data",
                                   fs_vfstype='nfs',
                                   fs_mntops="noauto,noatime,rsize=32768,wsize=32768",
                                   fs_freq="0", fs_passno="0")]
    current_actor_context.feed(StorageInfo(mount=with_mount_share,
                                           systemdmount=[],
                                           fstab=with_fstab_entry))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_not_skipped_if_initram_network_empty(current_actor_context, monkeypatch):
    """Check that previous inhibitors are not stopping the upgrade in case env var is set"""
    monkeypatch.setattr(config, 'get_env', lambda x, y: '' if x == 'LEAPP_DEVEL_INITRAM_NETWORK' else y)
    with_mount_share = [MountEntry(name="nfs", mount="/mnt/data", tp='nfs',
                                   options="rw,nosuid,nodev,relatime,user_id=1000,group_id=1000")]
    with_fstab_entry = [FstabEntry(fs_spec="lithium:/mnt/data", fs_file="/mnt/data",
                                   fs_vfstype='nfs',
                                   fs_mntops="noauto,noatime,rsize=32768,wsize=32768",
                                   fs_freq="0", fs_passno="0")]
    current_actor_context.feed(StorageInfo(mount=with_mount_share,
                                           systemdmount=[],
                                           fstab=with_fstab_entry))
    current_actor_context.run()
    report_fields = current_actor_context.consume(Report)[0].report
    assert is_inhibitor(report_fields)



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class MultipathConfigOption(Model):
    """Model representing information about a multipath configuration option"""
    topic = SystemInfoTopic

    name = fields.String(default='')
    value = fields.String(default='')


class MultipathConfig(Model):
    """Model representing information about a multipath configuration file"""
    topic = SystemInfoTopic

    pathname = fields.String()
    """Config file path name"""

    default_path_checker = fields.Nullable(fields.String())
    config_dir = fields.Nullable(fields.String())
    """Values of path_checker and config_dir in the defaults section.
       None if not set"""

    default_retain_hwhandler = fields.Nullable(fields.Boolean())
    default_detect_prio = fields.Nullable(fields.Boolean())
    default_detect_checker = fields.Nullable(fields.Boolean())
    reassign_maps = fields.Nullable(fields.Boolean())
    """True if retain_attached_hw_handler, detect_prio, detect_path_checker,
       or reassign_maps is set to "yes" in the defaults section. False
       if set to "no". None if not set."""

    hw_str_match_exists = fields.Boolean(default=False)
    ignore_new_boot_devs_exists = fields.Boolean(default=False)
    new_bindings_in_boot_exists = fields.Boolean(default=False)
    unpriv_sgio_exists = fields.Boolean(default=False)
    detect_path_checker_exists = fields.Boolean(default=False)
    overrides_hwhandler_exists = fields.Boolean(default=False)
    overrides_pg_timeout_exists = fields.Boolean(default=False)
    queue_if_no_path_exists = fields.Boolean(default=False)
    all_devs_section_exists = fields.Boolean(default=False)
    """True if hw_str_match, ignore_new_boot_devs, new_bindings_in_boot,
       detect_path_checker, or unpriv_sgio is set in any section,
       if queue_if_no_path is included in the features line in any
       section or if hardware_handler or pg_timeout is set in the
       overrides section. False otherwise"""

    all_devs_options = fields.List(fields.Model(MultipathConfigOption),
                                   default=[])
    """options in an all_devs device configuration section to be converted to
       an overrides section"""


class MultipathConfFacts(Model):
    """Model representing information from multipath configuration files"""
    topic = SystemInfoTopic

    configs = fields.List(fields.Model(MultipathConfig), default=[])
    """List of multipath configuration files"""



from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckGrep(Actor):
    """
    Check if Grep is installed. If yes, write information about non-compatible changes.
    """

    name = 'checkgrep'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(DistributionSignedRPM, 'grep'):
            create_report([
                reporting.Title('Grep has incompatible changes in the next major version'),
                reporting.Summary(
                    'If a file contains data improperly encoded for the current locale, and this is '
                    'discovered before any of the file\'s contents are output, grep now treats the file '
                    'as binary.\n'
                    'The \'grep -P\' no longer reports an error and exits when given invalid UTF-8 data. '
                    'Instead, it considers the data to be non-matching.\n'
                    'In locales with multibyte character encodings other than UTF-8, grep -P now reports '
                    'an error and exits instead of misbehaving.\n'
                    'When searching binary data, grep now may treat non-text bytes as line terminators. '
                    'This can boost performance significantly.\n'
                    'The \'grep -z\' no longer automatically treats the byte \'\\200\' as binary data.\n'
                    'Context no longer excludes selected lines omitted because of -m. For example, '
                    '\'grep "^" -m1 -A1\' now outputs the first two input lines, not just the first '
                    'line.\n'
                ),
                reporting.Severity(reporting.Severity.LOW),
                reporting.Groups([reporting.Groups.TOOLS]),
                reporting.Remediation(hint='Please update your scripts to be compatible with the changes.'),
                reporting.RelatedResource('package', 'grep')
            ])



from leapp.actors import Actor
from leapp.libraries.actor import scancustomrepofile
from leapp.models import CustomTargetRepository, CustomTargetRepositoryFile
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanCustomRepofile(Actor):
    """
    Scan the custom /etc/leapp/files/leapp_upgrade_repositories.repo repo file.

    This is the official path where to put the YUM/DNF repository file with
    custom repositories for the target system. These repositories will be used
    automatically for the in-place upgrade despite the enable/disable settings.

    Additionally the CustomTargetRepositoryFile message is produced if the file
    exists to let the other actors know they should handle the file as well.

    If the file doesn't exist, nothing happens.
    """

    name = 'scan_custom_repofile'
    consumes = ()
    produces = (CustomTargetRepository, CustomTargetRepositoryFile)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        scancustomrepofile.process()
